# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-06-30

from datetime import datetime, timedelta
from netCDF4 import Dataset, num2date, date2num
from poets.grid.grids import ShapeGrid, RegularGrid
from poets.image.resampling import resample_to_shape, average_layers
from poets.io.download import download_http, download_ftp, download_sftp, \
    get_file_date, download_local
from poets.io.fileformats import select_file
from poets.io.unpack import unpack, check_compressed
import math
import numpy as np
import os
import pandas as pd
import poets.grid.grids as gr
import poets.image.netcdf as nc
import poets.timedate.dateindex as dt
import shutil

class BasicSource(object):
    """Base Class for data sources.

    Parameters
    ----------
    name : str
        Name of the data source.
    filename : str
        Structure/convention of the file name.
    filedate : dict
        Position of date fields in filename, given as tuple.
    temp_res : str
        Temporal resolution of the source.
    rootpath : str
        Root path where all data will be stored.
    host : str
        Link to data host.
    protocol : str
        Protocol for data transfer.
    username : str, optional
        Username for data access.
    password : str, optional
        Password for data access.
    port : int, optional
        Port to data host, defaults to 22.
    directory : str, optional
        Path to data on host.
    dirstruct : list of strings, optional
        Structure of source directory, each list item represents a
        subdirectory.
    regions : list of str, optional
        List of regions where data from source is available. Uses all regions
        specified in dest_regions if not set.
    begin_date : datetime, optional
        Date from which on data is available.
    variables : string or list of strings, optional
        Variables used from data source, defaults to ['dataset'].
    nan_value : int, float, optional
        Nan value of the original data as given by the data provider.
    valid_range : tuple of int of float, optional
        Valid range of data, given as (minimum, maximum).
    data_range : tuple of int of float, optional
        Range of the values as data given in rawdata (minimum, maximum).
        Will be scaled to valid_range.
    ffilter : str, optional
        Pattern that apperas in filename. Can be used to select out not
        needed files if multiple files per date are provided.
    colorbar : str, optional
        Colorbar to use, use one from
        http://matplotlib.org/examples/color/colormaps_reference.html,
        defaults to jet.
    labels : list, optional
        Custom tick-labels for the legend in the web-app; must have same
        dimension as xticks and only works if xticks is set; Defaults
        to None.
    xticks : list of int or float, optional
        Custom tick locations for the legend in the web-app; must have same
        dimension as labels and only works if labels is set; Defaults
        to None.
    unit : str, optional
        Unit of dataset for displaying in legend. Does not have to be set
        if unit is specified in input file metadata. Defaults to None.
    dest_nan_value : int, float, optional
        NaN value in the final NetCDF file.
    dest_regions : list of str, optional
        Regions of interest where data should be resampled to.
    dest_sp_res : int, float, optional
        Spatial resolution of the destination NetCDF file, defaults to 0.25
        degree.
    dest_temp_res : string, optional
        Temporal resolution of the destination NetCDF file, possible values:
        ('day', 'week', 'dekad', 'month'), defaults to dekad.
    dest_start_date : datetime, optional
        Start date of the destination NetCDF file, defaults to 2000-01-01.
    src_file : dict of str, optional
        Path to file that contains source. Uses default NetCDF file if None.
        Key of dict must be regions as set in regions attribute.

    Attributes
    ----------
    name : str
        Name of the data source.
    filename : str
        Structure/convention of the file name.
    filedate : dict
        Position of date fields in filename, given as tuple.
    temp_res : str
        Temporal resolution of the source.
    host : str
        Link to data host.
    protocol : str
        Protocol for data transfer.
    username : str
        Username for data access.
    password : str
        Password for data access.
    port : int
        Port to data host.
    directory : str
        Path to data on host.
    dirstruct : list of strings
        Structure of source directory, each list item represents a
        subdirectory.
    regions : list of str
        List of regions where data from source is available.
    begin_date : datetime
        Date from which on data is available.
    ffilter : str
        Pattern that apperas in filename.
    colorbar : str, optional
        Colorbar to used.
    labels : list
        Custom tick-labels for the legend in the web-app.
    xticks : list of int or float
        Custom tick locations for the legend in the web-app.
    unit : str
        Unit of dataset for displaying in legend.
    variables : list of strings
        Variables used from data source.
    nan_value : int, float
        Not a number value of the original data as given by the data provider.
    valid_range : tuple of int of float
        Valid range of data, given as (minimum, maximum).
    data_range : tuple of int of float
        Range of the values as data given in rawdata (minimum, maximum).
    dest_nan_value : int, float, optional
        NaN value in the final NetCDF file.
    tmp_path : str
        Path where temporary files are stored.
    rawdata_path : str
        Path where original files are stored.
    data_path : str
        Path where resampled NetCDF file is stored.
    dest_regions : list of str
        Regions of interest where data is resampled to.
    dest_sp_res : int, float
        Spatial resolution of the destination NetCDF file.
    dest_temp_res : string
        Temporal resolution of the destination NetCDF file.
    dest_start_date : datetime.datetime
        First date of the dataset in the destination NetCDF file.
    src_file : str, list of str
        Path to file that contains source.
    """

    def __init__(self, name, filename, filedate, temp_res, rootpath,
                 host, protocol, username=None, password=None, port=22,
                 directory=None, dirstruct=None, regions=None,
                 begin_date=None, ffilter=None, colorbar='jet',
                 variables=None, nan_value=None, valid_range=None, unit=None,
                 dest_nan_value=-99, dest_regions=None, dest_sp_res=0.25,
                 dest_temp_res='dekad', dest_start_date=datetime(2000, 1, 1),
                 data_range=None, src_file=None, labels=None, xticks=None):

        self.name = name
        self.filename = filename
        self.filedate = filedate
        self.temp_res = temp_res
        self.host = host
        self.protocol = protocol
        self.username = username
        self.password = password
        self.port = port
        self.directory = directory
        self.dirstruct = dirstruct
        if begin_date is None:
            self.begin_date = dest_start_date
        else:
            self.begin_date = begin_date
        if type(variables) == str:
            self.variables = [variables]
        else:
            self.variables = variables
        self.ffilter = ffilter
        self.unit = unit
        self.nan_value = nan_value
        self.valid_range = valid_range
        self.data_range = data_range
        self.colorbar = colorbar
        self.labels = labels
        self.xticks = xticks

        if isinstance(regions, str):
            self.regions = [regions]
        else:
            self.regions = regions
        self.dest_nan_value = dest_nan_value
        if isinstance(dest_regions, str):
            self.dest_regions = [dest_regions]
        else:
            self.dest_regions = dest_regions
        self.dest_sp_res = dest_sp_res
        self.dest_temp_res = dest_temp_res
        self.dest_start_date = dest_start_date

        self.rawdata_path = os.path.join(rootpath, 'RAWDATA', name)
        self.tmp_path = os.path.join(rootpath, 'TMP')
        if not os.path.exists(self.tmp_path):
            os.mkdir(self.tmp_path)
        self.data_path = os.path.join(rootpath, 'DATA')
        if not os.path.exists(self.data_path):
            os.mkdir(self.data_path)

        if self.host[-1] != '/':
            self.host += '/'

        if self.directory is not None and self.directory[-1] != '/':
            self.directory += '/'

        if src_file is None:
            self.src_file = {}
            for reg in self.dest_regions:
                self.src_file[reg] = os.path.join(self.data_path, reg + '_' +
                                                  str(self.dest_sp_res) + '_'
                                                  + str(self.dest_temp_res)
                                                  + '.nc')
        else:
            self.src_file = src_file

    def _check_current_date(self, begin=True, end=True):
        """Helper method that checks the current date of individual variables
        in the netCDF data file.

        Parameters
        ----------
        begin : bool, optional
            If set True, begin will be returned as None.
        end : bool, optional
            If set True, end will be returned as None.
        Returns
        -------
        dates : dict of dicts
            Dictionary with dates of each parameter. None if no date available.
        """

        dates = {}

        for region in self.dest_regions:

            if self.regions is not None:
                if region not in self.regions:
                    continue

            nc_name = self.src_file[region]

            if os.path.exists(nc_name):
                dates[region] = {}

                variables = self.get_variables()

                with Dataset(nc_name, 'r', format='NETCDF4') as nc:
                    for var in variables:
                        dates[region][var] = []
                        if begin:
                            for i in range(0, nc.variables['time'].size - 1):
                                data = nc.variables[var][i]
                                if isinstance(data, np.ma.MaskedArray):
                                    if data.mask.min():
                                        continue
                                elif(math.isnan(np.nanmax(data))):
                                    continue

                                times = nc.variables['time']
                                dat = num2date(nc.variables['time'][i],
                                               units=times.units,
                                               calendar=times.calendar)
                                dates[region][var].append(dat)
                                break
                        else:
                            dates[region][var].append(None)

                        if end:
                            for i in range(nc.variables['time'].size - 1,
                                           - 1, -1):
                                data = nc.variables[var][i]
                                if isinstance(data, np.ma.MaskedArray):
                                    if data.mask.min():
                                        continue
                                elif(math.isnan(np.nanmax(data))):
                                    continue

                                times = nc.variables['time']
                                dat = num2date(nc.variables['time'][i],
                                               units=times.units,
                                               calendar=times.calendar)
                                dates[region][var].append(dat)
                                break
                        else:
                            dates[region][var].append(None)

                        if dates[region][var] in [[None], []]:
                            dates[region][var] = [None, None]
            else:
                dates = None
                break

        return dates

    def _get_download_date(self):
        """Gets the date from which to start the data download.

        Returns
        -------
        begin : datetime
            date from which to start the data download.
        """

        dates = self._check_current_date(begin=False)

        if dates is not None:
            begin = datetime.now()
            for region in self.dest_regions:

                if self.regions is not None:
                    if region not in self.regions:
                        continue

                variables = self.get_variables()
                if variables == []:
                    begin = self.dest_start_date
                else:
                    for var in variables:
                        if dates[region][var][1] is not None:
                            if dates[region][var][1] < begin:
                                begin = dates[region][var][1]
                                begin += timedelta(days=1)
                        else:
                            if self.dest_start_date < self.begin_date:
                                begin = self.begin_date
                            else:
                                begin = self.dest_start_date
        else:
            begin = self.begin_date

        return begin

    def _get_tmp_filepath(self, prefix, region):
        """Creates path to a temporary directory.

        Returns
        -------
        str
            Path to the temporary direcotry
        """
        filename = ('_' + prefix + '_' + region + '_' + str(self.dest_sp_res)
                    + '_' + str(self.dest_temp_res) + '.nc')
        return os.path.join(self.tmp_path, filename)

    def _resample_spatial(self, region, begin, end, delete_rawdata,
                          shapefile=None):
        """Helper method that calls spatial resampling routines.

        Parameters:
        region : str
            FIPS country code (https://en.wikipedia.org/wiki/FIPS_country_code)
        begin : datetime
            Start date of resampling
        end : datetime
            End date of resampling
        delete_rawdata : bool
            True if original downloaded files should be deleted after
            resampling
        """

        dest_file = self._get_tmp_filepath('spatial', region)

        dirList = os.listdir(self.rawdata_path)
        dirList.sort()

        if region == 'global':
            grid = gr.RegularGrid(sp_res=self.dest_sp_res)
        else:
            grid = gr.ShapeGrid(region, self.dest_sp_res, shapefile)

        for item in dirList:

            src_file = os.path.join(self.rawdata_path, item)

            fdate = get_file_date(item, self.filedate)

            if begin is not None:
                if fdate < begin:
                    continue

            if end is not None:
                if fdate > end:
                    continue

            if check_compressed(src_file):
                dirname = os.path.splitext(item)[0]
                dirpath = os.path.join(self.rawdata_path, dirname)
                unpack(src_file)
                src_file = select_file(os.listdir(dirpath))
                src_file = os.path.join(dirpath, src_file)

            if begin is not None:
                if fdate < begin:
                    if check_compressed(item):
                        shutil.rmtree(os.path.join(self.rawdata_path,
                                                   os.path.splitext(item)[0]))
                    continue
            if end is not None:
                if fdate > end:
                    if check_compressed(item):
                        shutil.rmtree(os.path.join(self.rawdata_path,
                                                   os.path.splitext(item)[0]))
                    continue

            print '.',

            image, _, _, _, timestamp, metadata = \
                resample_to_shape(src_file, region, self.dest_sp_res, grid,
                                  self.name, self.nan_value,
                                  self.dest_nan_value, self.variables,
                                  shapefile)

            if timestamp is None:
                timestamp = get_file_date(item, self.filedate)

            if self.temp_res == self.dest_temp_res:
                filename = (region + '_' + str(self.dest_sp_res) + '_'
                            + str(self.dest_temp_res) + '.nc')
                dfile = os.path.join(self.data_path, filename)
                nc.save_image(image, timestamp, region, metadata, dfile,
                              self.dest_start_date, self.dest_sp_res,
                              self.dest_nan_value, shapefile,
                              self.dest_temp_res)
            else:
                nc.write_tmp_file(image, timestamp, region, metadata,
                                  dest_file, self.dest_start_date,
                                  self.dest_sp_res, self.dest_nan_value,
                                  shapefile)

            # deletes unpacked files if existing
            if check_compressed(item):
                shutil.rmtree(os.path.join(self.rawdata_path,
                                           os.path.splitext(item)[0]))

        print ''

    def _resample_temporal(self, region, shapefile=None):
        """Helper method that calls temporal resampling routines.

        Parameters:
        region : str
            Identifier of the region in the shapefile. If the default shapefile
            is used, this would be the FIPS country code.

        shapefile : str, optional
            Path to shape file, uses "world country admin boundary shapefile"
            by default.
        """

        src_file = self._get_tmp_filepath('spatial', region)

        if not os.path.exists(src_file):
            print '[Info] No data available for this period'
            return False

        data = {}
        variables, _, period = nc.get_properties(src_file)

        dtindex = dt.get_dtindex(self.dest_temp_res, period[0], period[1])

        for date in dtindex:
            # skip if data for period is not complete
            # if date > period[1]:
            #    continue
            if self.dest_temp_res == 'dekad':
                if date.day < 21:
                    begin = datetime(date.year, date.month, date.day - 10 + 1)
                else:
                    begin = datetime(date.year, date.month, 21)
                end = date
            else:
                begin = period[0]
                end = date

            data = {}
            metadata = {}

            for var in variables:
                img, _, _, meta = \
                    nc.read_variable(src_file, var, begin, end)

                metadata[var] = meta
                data[var] = average_layers(img, self.dest_nan_value)

            dest_file = self.src_file[region]

            nc.save_image(data, date, region, metadata, dest_file,
                          self.dest_start_date, self.dest_sp_res,
                          self.dest_nan_value, shapefile, self.dest_temp_res)

        # delete intermediate netCDF file
        print ''
        os.unlink(src_file)

    def _scale_values(self, data):

        if self.valid_range is not None:
            if self.data_range is not None:
                data = ((data - self.data_range[0]) /
                        (self.data_range[1] - self.data_range[0]) *
                        (self.valid_range[1] - self.valid_range[0]) +
                        self.valid_range[0])

        return data

    def download(self, download_path=None, begin=None, end=None):
        """"Download data

        Parameters
        ----------
        begin : datetime, optional
            start date of download, default to None
        end : datetime, optional
            start date of download, default to None

        Returns
        -------
        check : bool, string
            True if download was complete, False if no data were available,
            Error Message if connection to resource failed.
        """

        errormsg = '[ERROR] Could not connect to source'

        if begin is None:
            if self.dest_start_date < self.begin_date:
                begin = self.begin_date
            else:
                begin = self.dest_start_date

        if self.protocol in ['HTTP', 'http']:
            try:
                check = download_http(self.rawdata_path, self.host,
                                      self.directory, self.filename,
                                      self.filedate, self.dirstruct,
                                      begin=begin, end=end,
                                      ffilter=self.ffilter)
            except:
                check = False
        elif self.protocol in ['FTP', 'ftp']:
            try:
                check = download_ftp(self.rawdata_path, self.host,
                                     self.directory, self.filedate, self.port,
                                     self.username, self.password,
                                     self.dirstruct, begin=begin, end=end,
                                     ffilter=self.ffilter)
            except:
                check = errormsg
        elif self.protocol in ['SFTP', 'sftp']:
            try:
                check = download_sftp(self.rawdata_path, self.host,
                                      self.directory, self.port, self.username,
                                      self.password, self.filedate,
                                      self.dirstruct, begin=begin, end=end,
                                      ffilter=self.ffilter)
            except:
                check = errormsg
        elif self.protocol in ['local', 'LOCAL']:
            try:
                check = download_local(self.rawdata_path, directory=self.host,
                                       filedate=self.filedate,
                                       dirstruct=self.dirstruct, begin=begin,
                                       end=end, ffilter=self.ffilter)
            except:
                check = errormsg

        return check

    def resample(self, begin=None, end=None, delete_rawdata=False,
                 shapefile=None, stepwise=True):
        """Resamples source data to given spatial and temporal resolution.

        Writes resampled images into a netCDF data file. Deletes original
        files if flag delete_rawdata is set True.

        Parameters
        ----------
        begin : datetime
            Start date of resampling.
        end : datetime
            End date of resampling.
        delete_rawdata : bool
            Original files will be deleted from rawdata_path if set 'True'.
        shapefile : str, optional
            Path to shape file, uses "world country admin boundary shapefile"
            by default.
        """

        if len(os.listdir(self.tmp_path)) != 0:
            for fname in os.listdir(self.tmp_path):
                if '.nc' in fname:
                    os.remove(os.path.join(self.tmp_path, fname))

        # clean rawdata folder from sudirectories
        for item in os.listdir(self.rawdata_path):
            if os.path.isdir(os.path.join(self.rawdata_path, item)):
                os.rmdir(os.path.join(self.rawdata_path, item))

        begin, end = self._check_begin_end(begin, end)

        if begin > end:
            print '[INFO] everything up to date'
            return '[INFO] everything up to date'

        if stepwise:

            drange = dt.get_dtindex(self.dest_temp_res, begin, end)

            for i, date in enumerate(drange):
                if i == 0:
                    start = begin
                else:
                    if self.dest_temp_res in ['dekad', 'dekadal', 'week',
                                              'weekly', 'month', 'monthly']:
                        start = drange[i - 1] + timedelta(days=1)
                    else:
                        start = date

                stop = date

                print '[INFO] Resampling ' + str(start) + ' to ' + str(stop)

                for region in self.dest_regions:

                    if self.regions is not None:
                        if region not in self.regions:
                            continue

                    print '[INFO] resampling to region ' + region
                    print '[INFO] performing spatial resampling ',

                    self._resample_spatial(region, start, stop, delete_rawdata,
                                           shapefile)

                    if self.temp_res == self.dest_temp_res:
                        print '[INFO] skipping temporal resampling'
                    else:
                        print '[INFO] performing temporal resampling ',
                        self._resample_temporal(region, shapefile)

        else:

            print '[INFO] ' + str(begin) + '-' + str(end)

            for region in self.dest_regions:

                if self.regions is not None:
                    if region not in self.regions:
                        continue

                print '[INFO] resampling to region ' + region
                print '[INFO] performing spatial resampling ',

                self._resample_spatial(region, begin, end, delete_rawdata,
                                       shapefile)

                if self.temp_res == self.dest_temp_res:
                    print '[INFO] skipping temporal resampling'
                else:
                    print '[INFO] performing temporal resampling ',
                    self._resample_temporal(region, shapefile)

        if delete_rawdata:
            print '[INFO] Cleaning up rawdata'
            dirList = os.listdir(self.rawdata_path)
            dirList.sort()

            for item in dirList:
                src_file = os.path.join(self.rawdata_path, item)
                os.unlink(src_file)

    def download_and_resample(self, download_path=None, begin=None, end=None,
                              delete_rawdata=False, shapefile=None):
        """Downloads and resamples data.

        Parameters
        ----------
        download_path : str
            Path where to save the downloaded files.
        begin : datetime.date, optional
            set either to first date of remote repository or date of
            last file in local repository
        end : datetime.date, optional
            set to today if none given
        delete_rawdata : bool, optional
            Original files will be deleted from rawdata_path if set True
        shapefile : str, optional
            Path to shape file, uses "world country admin boundary shapefile"
            by default.
        """

        begin, end = self._check_begin_end(begin, end)

        if begin > end:
            print '[INFO] everything up to date'
            return '[INFO] everything up to date'

        drange = dt.get_dtindex(self.dest_temp_res, begin, end)

        for i, date in enumerate(drange):
            if i == 0:
                start = begin
            else:
                if self.dest_temp_res in ['dekad', 'dekadal', 'week',
                                          'weekly', 'month', 'monthly']:
                    start = drange[i - 1] + timedelta(days=1)
                else:
                    start = date

            stop = date

            filecheck = self.download(download_path, start, stop)
            if filecheck is True:
                self.resample(start, stop, delete_rawdata, shapefile, False)
            else:
                if filecheck is False:
                    print '[WARNING] no data available for this date'
                else:
                    print filecheck

    def read_ts(self, location, region=None, variable=None, shapefile=None,
                scaled=True):
        """Gets timeseries from netCDF file for a gridpoint.

        Parameters
        ----------
        location : int or tuple of floats
            Either Grid point index as integer value or Longitude/Latitude
            given as tuple.
        region : str, optional
            Region of interest, set to first defined region if not set.
        variable : str, optional
            Variable to display, selects all available variables if None.
        shapefile : str, optional
            Path to custom shapefile.
        scaled : bool, optional
            If true, data will be scaled to a predefined range; if false, data
            will be shown as given in rawdata file; defaults to True

        Returns
        -------
        df : pd.DataFrame
            Timeseries for selected variables.
        """

        if region is None:
            region = self.dest_regions[0]

        if type(location) is tuple:
            if region == 'global':
                grid = RegularGrid(self.dest_sp_res)
            else:
                grid = ShapeGrid(region, self.dest_sp_res, shapefile)

            gp, _ = grid.find_nearest_gpi(location[0], location[1])
        else:
            gp = location

        if variable is None:
            variable = self.get_variables()
        else:
            variable = self.check_variable(variable)
            variable = [variable]

        source_file = self.src_file[region]

        var_dates = self._check_current_date()

        with Dataset(source_file, 'r', format='NETCDF4') as nc:

            time = nc.variables['time']
            dates = num2date(time[:], units=time.units, calendar=time.calendar)
            position = np.where(nc.variables['gpi'][:] == gp)
            lat_pos = position[0][0]
            lon_pos = position[1][0]
            df = pd.DataFrame(index=pd.DatetimeIndex(dates))

            for ncvar in variable:
                begin = np.where(dates == var_dates[region][ncvar][0])[0][0]
                end = np.where(dates == var_dates[region][ncvar][1])[0][0]
                df[ncvar] = np.NAN
                for i in range(begin, end + 1):
                    df[ncvar][i] = nc.variables[ncvar][i, lat_pos, lon_pos]

                if 'scaling_factor' in nc.variables[ncvar].ncattrs():
                    vvar = nc.variables[ncvar]
                    if vvar.getncattr('scaling_factor') < 0:
                        df[ncvar] = (df[ncvar] *
                                     float(vvar.getncattr('scaling_factor')))
                    else:
                        df[ncvar] = (df[ncvar] /
                                     float(vvar.getncattr('scaling_factor')))
                if scaled:
                    if self.valid_range is not None:
                        if self.data_range is not None:
                            df[ncvar] = self._scale_values(df[ncvar])

        return df

    def read_img(self, date, region=None, variable=None, scaled=True):
        """Gets images from netCDF file for certain date

        Parameters
        ----------
        date : datetime
            Date of the image.
        region : str, optional
            Region of interest, set to first defined region if not set.
        variable : str, optional
            Variable to display, selects first available variables if None.
        scaled : bool, optional
            If true, data will be scaled to a predefined range; if false, data
            will be shown as given in rawdata file; defaults to True.

        Returns
        -------
        img : numpy.ndarray
            Image of selected date.
        lon : numpy.array
            Array with longitudes.
        lat : numpy.array
            Array with latitudes.
        metadata : dict
            Dictionary containing metadata of the variable.
        """

        if region is None:
            region = self.dest_regions[0]

        if variable is None:
            variable = self.get_variables()[0]
        else:
            # Renames variable name to SOURCE_variable
            variable = self.check_variable(variable)

        source_file = self.src_file[region]

        # get dekad of date:
        date = dt.check_period(self.dest_temp_res, date)

        with Dataset(source_file, 'r', format='NETCDF4') as nc:
            time = nc.variables['time']

            datenum = date2num(date, units=time.units, calendar=time.calendar)

            position = np.where(time[:] == datenum)[0][0]

            var = nc.variables[variable]
            img = var[position]
            lon = nc.variables['lon'][:]
            lat = nc.variables['lat'][:]

            metadata = {}

            for attr in var.ncattrs():
                if attr[0] != '_' and attr != 'scale_factor':
                    metadata[attr] = var.getncattr(attr)

            if not metadata:
                metadata = None

            if 'scaling_factor' in var.ncattrs():
                if metadata['scaling_factor'] < 0:
                    img = img * float(metadata['scaling_factor'])
                else:
                    img = img / float(metadata['scaling_factor'])

        if scaled:
            if self.valid_range is not None:
                if self.data_range is not None:
                    img = self._scale_values(img)

        return img, lon, lat, metadata

    def get_variables(self):
        """
        Gets all variables given in the NetCDF file.

        Returns
        -------
        variables : list of str
            Variables from given in the NetCDF file.
        """

        # nc_name = self.src_file[self.dest_regions[0]]

        # nc_vars, _, _ = nc.get_properties(nc_name)

        nc_vars = []
        for reg in self.dest_regions:
            vari, _, _ = nc.get_properties(self.src_file[reg])
            if vari is None:
                continue
            for v in vari:
                if v not in nc_vars:
                    nc_vars.append(v)

        variables = []

        if self.variables is not None:
            for var in self.variables:
                if var in nc_vars:
                    variables.append(var)
                else:
                    if self.name + '_' + var in nc_vars:
                        variables.append(self.name + '_' + var)
        else:
            for var in nc_vars:
                if self.name + '_dataset' in var:
                    variables.append(var)
                elif self.name in var:
                    variables.append(var)

        return variables

    def check_variable(self, variable):
        """
        Checks if a variable exists in a source and returns it's correct name.

        Parameters
        ----------
        variable : str
            Variable to check.

        Returns
        -------
        varname : str
            Name of the variable in the source.
        """

        varname = ''

        if self.variables is not None:
            for var in self.variables:
                if variable in var:
                    varname = variable
                    break
                elif self.name + '_' + variable in var:
                    varname = self.name + '_' + variable
                    break
                else:
                    if variable == self.name + '_' + var:
                        varname = self.name + '_' + var
                        break

        else:
            for var in self.get_variables():
                if variable == var:
                    varname = variable
                    break
                else:
                    if self.name + '_' + variable in var:
                        varname = self.name + '_' + variable
                        break

        return varname

    def _check_begin_end(self, begin, end):
        """
        Checks begin and end date and returns valid dates.

        Parameters
        ----------
        begin : datetime
            Begin date to check.
        end : datetime
            End date to check.

        Returns
        -------
        begin : datetime
            Begin date.
        end : datetime
            End date.
        """

        if begin is None:
            if self.dest_start_date < self.begin_date:
                begin = self.begin_date
            else:
                begin = self.dest_start_date

            if begin < self._get_download_date():
                begin = self._get_download_date()

            # start one period earlier to close possible gaps
            begin = begin - timedelta(days=1)
            begin, _ = dt.check_period_boundaries(self.dest_temp_res, begin)
            if begin < self.begin_date:
                begin = self.begin_date
            if begin < self.dest_start_date:
                begin = self.dest_start_date

        if end is None:
            end = datetime.now()

        return begin, end

    def fill_gaps(self, begin=None, end=None):
        """
        Detects gaps in data and tries to fill them by downloading and
        resampling the data within these periods.

        Parameters
        ----------
        begin : datetime
            Begin date of intervall to check, defaults to None.
        end : datetime
            End date of intervall to check, defaults to None.
        """

        gaps = []

        for region in self.dest_regions:

            if self.regions is not None:
                if region not in self.regions:
                    continue

            _, _, period = nc.get_properties(self.src_file[region])
            if begin is None:
                if self.begin_date < self.dest_start_date:
                    begin = self.dest_start_date
                else:
                    begin = self.begin_date
            if end is None:
                end = period[1]
            drange = dt.get_dtindex(self.dest_temp_res, begin, end)
            for date in drange:
                nonans = []
                for var in self.get_variables():
                    img, _, _, _ = self.read_img(date, region, var)
                    if np.nanmean(img) is not np.ma.masked:
                        nonans.append(1)
                if len(nonans) == 0:
                    if date not in gaps:
                        gaps.append(date)

        if len(gaps) == 0:
            print '[INFO] No gaps found.'
        else:
            print('[INFO] Found ' + str(len(gaps)) +
                  ' gap(s), attempt to fill..')
            for date in gaps:
                if self.dest_temp_res in ['day', 'daily']:
                    begin = date
                    end = date
                else:
                    begin, end = dt.check_period_boundaries(self.dest_temp_res,
                                                            date)

                self.download_and_resample(begin=begin, end=end)



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-05-26

from cStringIO import StringIO
import os
from flask import Flask, render_template, jsonify, make_response
from flask.ext.cors import CORS
import numpy as np
import pandas as pd
from poets.timedate.dateindex import get_dtindex
from poets.web.overlays import image_bounds
import pytesmo.time_series as ts
import urlparse
import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.use('Agg')


def curpath():
    """
    Gets the current path of the module.

    Returns
    -------
    pth : str
        Path of the module.
    """
    pth, _ = os.path.split(os.path.abspath(__file__))
    return pth


def to_dygraph_format(self):
    """
    Transforms pandas DataFrame to Dygraphs compatible format.

    Returns
    -------
    labels : list of str
        Labels of the Dygraphs array.
    values : list
        Values of the Dygraphs array.
    """

    labels = ['date']
    labels.extend(self.columns.values.tolist())
    data_values = np.hsplit(self.values, self.columns.values.size)
    data_index = self.index.values.astype('M8[s]').tolist()
    data_index = [x.strftime("%Y/%m/%d %H:%M:%S") for x in data_index]
    data_index = np.reshape(data_index, (len(data_index), 1))
    data_values.insert(0, data_index)
    data_values = np.column_stack(data_values)
    values = data_values.tolist()

    return labels, values

pd.DataFrame.to_dygraph_format = to_dygraph_format

# dest = os.path.join(curpath(), 'static', 'temp')


class ReverseProxied(object):
    '''Wrap the application in this middleware and configure the
    front-end server to add these headers, to let you quietly bind
    this to a URL other than / and to an HTTP scheme that is
    different than what is used locally.

    In nginx:
    location /myprefix {
        proxy_pass http://192.168.0.1:5001;
        proxy_set_header Host $host;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-Scheme $scheme;
        proxy_set_header X-Script-Name /myprefix;
        }

    :param app: the WSGI application
    '''
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get('HTTP_X_SCRIPT_NAME', '')
        if script_name:
            environ['SCRIPT_NAME'] = script_name
            path_info = environ['PATH_INFO']
            if path_info.startswith(script_name):
                environ['PATH_INFO'] = path_info[len(script_name):]

        scheme = environ.get('HTTP_X_SCHEME', '')
        if scheme:
            environ['wsgi.url_scheme'] = scheme
        return self.app(environ, start_response)

app = Flask(__name__, static_folder='static', static_url_path='/static',
            template_folder="templates")
app.config['CORS_HEADERS'] = 'Content-Type'
# app.wsgi_app = ReverseProxied(app.wsgi_app)

cors = CORS(app, resources={r"/*": {"origins": "*"}})


def start(poet, host='127.0.0.1', port=None, r_host=None, r_port=None,
          url=None, debug=False):
    """
    Starts application and sets global variables.

    Parameters
    ----------
    poet : Poet()
        Instance of Poet class.
    host : str, optional
        Host that is used by the app, defaults to 127.0.0.1.
    port : int, optional
        Port where app runs on, defaults to 50000.
    r_host : str, optional
        IP of router that is between host and internet.
    r_port : int, optional
        Port of router that is between host and internet.
    debug : bool, optional
        Starts app in debug mode if set True, defaults to False.
    """

    global p
    global variables
    global dates
    global vmin, vmax, cmap
    global host_gl
    global port_gl
    global url_gl

    p = poet
    variables = poet.get_variables()

    if port is None:
        port = 5000

    if r_host is None:
        host_gl = host
    else:
        host_gl = r_host

    if r_port is None:
        port_gl = port
    else:
        port_gl = r_port

    if url is not None:
        pa = urlparse.urlparse(url, 'http')
        url_gl = os.path.join(pa.geturl(), '')
    else:
        url_gl = url

    if debug:
        app.run(debug=True, use_debugger=True, use_reloader=True, host=host,
                port=port)
    else:
        app.run(host=host, port=port)


@app.route('/', methods=['GET', 'POST'])
@app.route('/<reg>&<var>', methods=['GET', 'POST'])
def index(**kwargs):
    """
    Renders main page of the web application. Generates image arguments needed
    for OpenLayers overlay if parameters `reg` and `var` are set, renders
    entry page if not set.
    """

    global enddate
    global dates
    global ndate

    regions = []
    for i, reg in enumerate(p.regions):
        regions.append({'code': reg, 'name': p.region_names[i]})

    if len(kwargs) > 0:

        if 'reg' in kwargs:
            region = kwargs['reg']
        if 'var' in kwargs:
            variable = kwargs['var']

        for src in p.sources.keys():
            if variable in p.sources[src].get_variables():
                source = p.sources[src]

        ndate = source._check_current_date()
        begindate = ndate[region][variable][0]
        enddate = ndate[region][variable][1]

        if begindate is None and enddate is None:

            error = 'No data available for this dataset.'

            return render_template('index.html',
                                   regions=p.regions,
                                   sources=p.sources.keys(),
                                   variables=variables,
                                   error=error)

        d = get_dtindex(p.temporal_resolution, begindate, enddate)
        dates = d.to_pydatetime()

        fdates = []

        for i, d in enumerate(dates.tolist()):
            dat = {'id': i, 'date': d.strftime('%Y-%m-%d')}
            fdates.append(dat)

        lon_min, lon_max, lat_min, lat_max, c_lat, c_lon, _ = \
            image_bounds(region, p.spatial_resolution, p.shapefile)

        if source.valid_range is None:
            vrange = [-999, -999]
        else:
            vrange = source.valid_range

        ex1 = (lon_max, lat_min)
        ex2 = (lon_min, lat_max)

        return render_template('app.html',
                               max=len(dates) - 1,
                               coord=[c_lon, c_lat],
                               ex1=ex1,
                               ex2=ex2,
                               region=region,
                               source=source.name,
                               variable=variable,
                               regions=regions,
                               variables=variables,
                               dates=fdates,
                               host=host_gl,
                               port=port_gl,
                               sp_res=p.spatial_resolution,
                               range=vrange,
                               url=url_gl)
    else:
        return render_template('index.html',
                               regions=regions,
                               sources=p.sources.keys(),
                               variables=variables,
                               host=host_gl,
                               port=port_gl,
                               url=url_gl)


@app.route('/_ts/<reg>&<src>&<var>&<loc>', methods=['GET', 'OPTIONS'])
@app.route('/_ts/<reg>&<src>&<var>&<loc>&<anom>', methods=['GET', 'OPTIONS'])
def get_ts(**kwargs):
    """
    Gets time series for selected location, gets anomaly of time series if
    `anom` parameter is passed.

    Returns
    -------
    jsonified str
        Time series (anomaly) in Dygraphs compatible json format.
    """

    anomaly = False

    if 'reg' in kwargs:
        region = kwargs['reg']
    if 'src' in kwargs:
        source = p.sources[kwargs['src']].name
    if 'var' in kwargs:
        variable = kwargs['var']
    if 'loc' in kwargs:
        loc = kwargs['loc']
    if 'anom' in kwargs:
        anomaly = True

    loc = loc.split(',')
    lonlat = (float(loc[0]), float(loc[1]))

    df = p.read_timeseries(source, lonlat, region, variable)

    if anomaly:
        df = ts.anomaly.calc_anomaly(df, window_size=100)
        columns = []
        for cols in df.columns:
            columns.append(cols + '_anomaly')
        df.columns = columns

    labels, values = df.to_dygraph_format()
    data = {'labels': labels, 'data': values}

    return jsonify(data)


@app.route('/_tsdown/<reg>&<src>&<var>&<loc>')
@app.route('/_tsdown/<reg>&<src>&<var>&<loc>&<anom>')
def download_ts(**kwargs):
    """
    Initiates download time series (anomaly) in comma separated values format.

    Returns
    -------
    jsonified str
        Time series (anomaly) in Dygraphs compatible json format.
    """

    anomaly = False

    if 'reg' in kwargs:
        region = kwargs['reg']
    if 'src' in kwargs:
        source = p.sources[kwargs['src']]
    if 'var' in kwargs:
        variable = kwargs['var']
    if 'loc' in kwargs:
        loc = kwargs['loc']
    if 'anom' in kwargs:
        anomaly = True

    loc = loc.split(',')
    lonlat = (float(loc[0]), float(loc[1]))

    filename = region + '_' + variable + '_' + loc[0][:6] + '_' + loc[1][:6]

    df = p.read_timeseries(source.name, lonlat, region, variable)

    if anomaly:
        df = ts.anomaly.calc_anomaly(df)
        columns = []
        for cols in df.columns:
            columns.append(cols + '_anomaly')
        df.columns = columns

        filename += '_anomaly'

    output = StringIO()

    df.to_csv(output)
    csv = output.getvalue()

    response = make_response(csv)
    response.headers["Content-Disposition"] = ("attachment; filename=" +
                                               filename + ".csv")

    return response


@app.route('/_rimg/<reg>&<src>&<var>&<idx>', methods=['GET', 'POST'])
def request_image(**kwargs):
    """
    Creates image for OpenLayers overlay.

    Returns
    -------
    StringIO
        Image in StringIO.
    """

    global vmin
    global vmax
    global metadata
    global cmap

    if 'reg' in kwargs:
        region = kwargs['reg']
    if 'src' in kwargs:
        source = p.sources[kwargs['src']]
    if 'var' in kwargs:
        variable = kwargs['var']
    if 'idx' in kwargs:
        idx = kwargs['idx']

    pidx = (dates[int(idx)])

    img, _, _, metadata = p.read_image(source.name, pidx, region, variable)

    if source.unit is not None:
        if metadata is not None and 'unit' not in metadata:
            metadata['unit'] = source.unit
        elif metadata is None:
            metadata = {}
            metadata['unit'] = source.unit

    if source.labels is not None and source.xticks is not None:
        if metadata is None:
            metadata = {}
        metadata['labels'] = source.labels
        metadata['xticks'] = source.xticks

    if source.valid_range is not None:
        vmin = source.valid_range[0]
        vmax = source.valid_range[1]
    else:
        vmin = np.nanmin(img)
        vmax = np.nanmax(img)

    cmap = source.colorbar

    # Rescale the image
    n = 10
    img = np.kron(img, np.ones((n, n)))
    img[img == p.nan_value] = np.NAN

    buf = StringIO()
    plt.imsave(buf, img, vmin=vmin, vmax=vmax, cmap=cmap)

    image = buf.getvalue()

    response = make_response(image)
    response.headers["Content-Type"] = ("image/png; filename=data.png")

    return response


@app.route('/_rlegend/<reg>&<var>', methods=['GET', 'POST'])
@app.route('/_rlegend/<reg>&<var>&<idx>', methods=['GET', 'POST'])
def request_legend(**kwargs):
    """
    Creates Legend for OpenLayers overlay.

    Returns
    -------
    StringIO
        Legend in StringIO.
    """

    global vmin
    global vmax
    global metadata
    global cmap

    fig = plt.figure(figsize=(4, 0.7))
    ax1 = fig.add_axes([0.05, 0.7, 0.9, 0.10])
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    cb1 = mpl.colorbar.ColorbarBase(ax1, cmap=cmap, norm=norm,
                                    orientation='horizontal')
    plt.xticks(fontsize=9)

    if metadata:
        units = ['units', 'unit', 'UNITS', 'UNIT']
        for unit in units:
            if unit in metadata:
                cb1.set_label(metadata[unit], fontsize=10)
        if 'labels' in metadata.keys() and 'xticks' in metadata.keys():
            cb1.set_ticks(metadata['xticks'])
            cb1.set_ticklabels(metadata['labels'])

    fig.patch.set_alpha(0.6)

    buf = StringIO()
    plt.savefig(buf)
    plt.close()

    image = buf.getvalue()

    response = make_response(image)
    response.headers["Content-Type"] = ("image/png; filename=legend.png")

    return response


@app.route('/_variables', methods=['GET', 'POST'])
@app.route('/_variables/', methods=['GET', 'POST'])
@app.route('/_variables/<reg>', methods=['GET', 'POST'])
def request_variables(**kwargs):

    if 'reg' in kwargs:
        region = kwargs['reg']
    else:
        region = None

    variables = {}
    variables['variables'] = p.get_variables(region)

    return jsonify(variables)


@app.route('/about')
def about():
    """
    Creates the `about` page.
    """
    return render_template('about.html', url=url_gl)



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-10-02

"""
This module provides functions for loading from HDF5 files.
"""

import h5py
import numpy as np


def read_image(source_file, variables=None):
    """Reads data out of hdf5 file and returns it as numpy.ndarray

    Parameters
    ----------
    source_file : str
        Path to source file.
    variables : list of str, optional
        Variables to read from file, reads all variables if not set.

    Returns
    -------
    data : dict of numpy.arrays
        Source file.
    lon : numpy.array
        Longitudes of the source file.
    lat : numpy.array
        Latitudes of the source file.
    timestamp : datetime.date
        Timestamp of image.
    metadata : dict of strings
        Metadata from source netCDF file.
    """

    with h5py.File(source_file, 'r') as h5:

        # extract time!!!
        timestamp = None

        data = {}
        metadata = {}
        meta_global = {}

        for attr in h5.attrs:
            if (isinstance(h5.attrs[attr], list) or
                isinstance(h5.attrs[attr], dict) or
                isinstance(h5.attrs[attr], tuple)):
                meta_global[str(attr).lower()] = h5.attrs[attr][0]
            else:
                meta_global[str(attr).lower()] = h5.attrs[attr]

        for grp in h5:
            for var in h5[grp].keys():
                if ((variables is not None and var in variables)
                    or variables is None):
                    metadata[str(var)] = {}
                    data[var] = h5[grp][var].value
                    for attr in meta_global:
                        metadata[var][attr] = meta_global[attr]
                    for attr in h5[grp][var].attrs:
                        if (isinstance(h5[grp][var].attrs[attr], list) or
                            isinstance(h5[grp][var].attrs[attr], dict) or
                            isinstance(h5[grp][var].attrs[attr], tuple)):
                            attr_value = h5[grp][var].attrs[attr][0]
                        else:
                            attr_value = h5[grp][var].attrs[attr]
                        metadata[var][str(attr)] = attr_value
                        if str(attr).lower() == 'scaling_factor':
                            metadata[var]['scale_factor'] = \
                                attr_value

        sp_res = 180. / data[data.keys()[0]].shape[0]

        lonmin = -180. + (sp_res / 2.)
        latmin = -90. + (sp_res / 2.)

        lon = np.arange(lonmin, 180, sp_res)
        lat = np.arange(latmin, 90, sp_res)

        lat = lat[::-1]

    return data, lon, lat, timestamp, metadata



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-08-14

import unittest
import os
import poets.image.imagefile as imf
import numpy as np
import numpy.testing as nptest
from PIL import Image


def curpath():
    pth, _ = os.path.split(os.path.abspath(__file__))
    return pth


class Test(unittest.TestCase):

    def setUp(self):
        self.testfile = os.path.join(curpath(), 'data', 'test')
        self.img = Image.open(os.path.join(self.testfile + '.png'))
        self.region = ['NZ', 'US', 'RS', 'IT']
        self.fileExtension = ['.png', '.tif']

        self.lon_array = [np.array([165.5, 166.5, 167.5, 168.5, 169.5, 170.5,
                                   171.5, 172.5, 173.5, 174.5, 175.5, 176.5,
                                   177.5, 178.5, 179.5, -179.5, -178.5, -177.5,
                                   - 176.5, -175.5]), np.array([]),
                          np.array([]), np.array([6.5, 7.5, 8.5, 9.5,
                                                  10.5, 11.5, 12.5, 13.5,
                                                  14.5, 15.5, 16.5, 17.5,
                                                  18.5])]

        self.lat_array = [np.array([-34.5, -35.5, -36.5, -37.5, -38.5, -39.5,
                                   - 40.5, -41.5, -42.5, -43.5, -44.5, -45.5,
                                   - 46.5, -47.5, -48.5, -49.5, -50.5, -51.5,
                                   - 52.5]), np.array([]), np.array([]),
                          np.array([47.5, 46.5, 45.5, 44.5, 43.5, 42.5,
                                    41.5, 40.5, 39.5, 38.5, 37.5, 36.5])]

        self.row_array = np.array([10, 20, 30, 40])
        self.col_array = np.array([10, 20, 30, 40])

        self.lon = 170
        self.lat = -45

        self.row = 135.0
        self.col = 350.0

        self.row_rearr = 135.0
        self.col_rearr = 170.0

    def tearDown(self):
        pass

    def test_get_layer_extent(self):
            lon_min, lat_min, lon_max, lat_max = \
                imf.get_layer_extent(os.path.join(self.testfile + '.png'))

            assert (lon_min, lon_max) == (0.0, 360.0)
            assert (lat_min, lat_max) == (180.0, 0.0)

    def test_bbox_img(self):

        datamean = [244.88947368421051, [], [], 224.75]
        datamin = [112, [], [], 141]
        datamax = [255, [], [], 255]

        for fExt in self.fileExtension:
            for ind in [0, 3]:
                if fExt == '.tif':
                    datamean = [0.94736842105263153, [], [],
                                     5.2756410256410255]
                    datamin = [0, [], [], 0]
                    datamax = [12, [], [], 11]

                data, lon_new, lat_new, _, _ = imf.bbox_img(os.path.join(self\
                                                            .testfile + fExt),
                                                            self.region[ind],
                                                            fExt)

                nptest.assert_array_equal(lon_new, self.lon_array[ind])
                nptest.assert_array_equal(lat_new, self.lat_array[ind])
                assert data['dataset'].mean() == datamean[ind]
                assert data['dataset'].min() == datamin[ind]
                assert data['dataset'].max() == datamax[ind]

    def test_lonlat2px(self):

        row, col = imf.lonlat2px(self.img, self.lon, self.lat)

        assert row == self.row
        assert col == self.col

    def test_lonlat2px_rearr(self):

        row, col = imf.lonlat2px_rearr(self.img, self.lon, self.lat)

        assert row == self.row_rearr
        assert col == self.col_rearr

    def test_px2lonlat(self):
        lon1 = np.array([-170., -160., -150., -140.])
        lat1 = np.array([80., 70., 60., 50.])
        lon, lat = imf.px2lonlat(self.img, self.col_array, self.row_array)

        nptest.assert_array_equal(lon1, lon)
        nptest.assert_array_equal(lat1, lat)

    def test_px2lonlat_rearr(self):
        lon1 = np.array([10, 20, 30, 40])
        lat1 = np.array([80., 70., 60., 50.])
        lon, lat = imf.px2lonlat_rearr(self.img, self.col_array,
                                       self.row_array)

        nptest.assert_array_equal(lon1, lon)
        nptest.assert_array_equal(lat1, lat)

    def test_rearrange_img(self):
        greyval = 139
        img_rearr = imf.rearrange_img(self.img)

        assert greyval == img_rearr.getpixel((self.col_rearr, self.row_rearr))

    def test_dateline_country(self):
        lon_min1 = [165.87, 173.18333333333334, 19.633333333333333]
        lon_max1 = [-175.83333333333334, -66.98353057222222, -169.065]

        for ind, region in enumerate(self.region[0:-1]):
            lon_min, lon_max = imf.dateline_country(region)

            assert lon_min == lon_min1[ind]
            assert lon_max == lon_max1[ind]

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Isabella Pfeil, isy.pfeil@gmx.at
# Creation date: 2014-09-11

"""
GeoTIFF operations.
"""

import numpy as np


def lonlat2px_gt(img, lon, lat, lon_min, lat_min, lon_max, lat_max):
    """
    Converts a pair of lon and lat to its corresponding pixel value in an
    geotiff image file.

    Parameters
    ----------
    img : Image File, e.g. PNG, TIFF
        Input image file
    lon : float
        Longitude
    lat : float
        Latitude
    lon_min, lat_min : float
        lower left coordinate of geotiff
    lon_max, lat_max : float
        upper right coordinate of geotiff

    Returns
    -------
    Row : float
        corresponding pixel value
    Col : float
        corresponding pixel value
    """

    w, h = img.size

    londiff = lon_max - lon_min
    latdiff = lat_max - lat_min

    mw = w / londiff
    mh = h / latdiff

    row = (-lat + lat_max) * mh
    col = (lon - lon_min) * mw

    return row, col


def px2lonlat_gt(img, lon_px, lat_px, lon_min, lat_min, lon_max, lat_max):
    """
    Converts two arrays of row and column pixels into their corresponding
    longitude and latitude arrays

    Parameters
    ----------
    img : Image file
        Image which the pixel values refer to
    lon_px : np.array
        array of column pixels
    lat_px : np.array
        array of row pixels

    Returns
    -------
    lon_new : np.array
        List of corresponding longitude values
    lat_new: np.array
        List of corresponding latitude values
    lon_min, lat_min : float
        lower left coordinate of geotiff
    lon_max, lat_max : float
        upper right coordinate of geotiff
    """

    w, h = img.size

    londiff = lon_max - lon_min
    latdiff = lat_max - lat_min

    mw = w / londiff
    mh = h / latdiff

    lon_new = np.zeros(len(lon_px))
    lat_new = np.zeros(len(lat_px))

    for i in range(0, len(lon_px)):
        lon_new[i] = lon_min + lon_px[i] / mw

    for i in range(0, len(lat_px)):
        lat_new[i] = lat_max - lat_px[i] / mh

    return lon_new, lat_new



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-07-08

import os
import unittest
from poets.shape.shapes import Shape, FipsError


def curpath():
    pth, _ = os.path.split(os.path.abspath(__file__))
    return pth


class Test(unittest.TestCase):

    def setUp(self):
        self.region = 'AU'
        self.name = 'Austria'

        self.shapefile = os.path.join(curpath(), 'data', 'shape', 'noe_shp')

    def tearDown(self):
        pass

    def test_Shape(self):

        # test default shapefile
        shp = Shape(self.region)

        assert shp.code == self.region
        assert shp.name == self.name
        assert shp.bbox == (9.5335693359375, 46.407493591308594,
                            17.166385650634766, 49.01874542236328)

        # test custom shapefile
        cshp = Shape('NOE', shapefile=self.shapefile)
        assert cshp.code == 'NOE'
        assert cshp.name == 'NOE'
        assert cshp.bbox == (14.455012306812042, 47.422602087551724,
                             17.07012761362633, 49.021162769139316)

        # test if error is raised
        self.assertRaises(FipsError, Shape, 'XYZ')

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-08-14

from datetime import datetime
import os
import shutil
import unittest

import numpy as np
from poets.io.source_base import BasicSource
from poets.timedate.dateindex import get_dtindex


def curpath():
    pth, _ = os.path.split(os.path.abspath(__file__))
    return pth


class Test(unittest.TestCase):

    @classmethod
    def setUpClass(self):

        self.rootpath = os.path.join(curpath(), 'data', 'src_test')
        self.dest_nan_value = -99
        self.dest_regions = ['AU']
        self.dest_sp_res = 1
        self.dest_temp_res = 'daily'
        self.dest_start_date = datetime(2013, 1, 7)
        self.enddate = datetime(2013, 1, 13)
        self.testdate = datetime(2013, 1, 10)
        self.name = 'TEST'
        self.filename = "test_{YYYY}_{MM}_{TT}.png"
        self.filedate = {'YYYY': (5, 9), 'MM': (10, 12), 'DD': (13, 15)}
        self.temp_res = 'daily'
        self.protocol = 'local'
        self.begin_date = datetime(2013, 1, 7)

        if os.path.exists(self.rootpath):
            shutil.rmtree(self.rootpath)

        os.mkdir(self.rootpath)
        os.mkdir(os.path.join(self.rootpath, 'DATA'))
        os.mkdir(os.path.join(self.rootpath, 'TMP'))

        # setup test png files
        self.pngdir = os.path.join(curpath(), 'data', 'testpngs')
        self.host = self.pngdir

        if os.path.exists(self.pngdir):
            shutil.rmtree(self.pngdir)
        os.mkdir(self.pngdir)
        self.dtindex = get_dtindex('day', self.dest_start_date,
                                   self.enddate)
        for i, dat in enumerate(self.dtindex):
            if i == 3:
                continue
            year = str(dat.year)
            month = "%02d" % (dat.month)
            day = "%02d" % (dat.day)
            fname = ('test_' + year + '_' + month + '_' + day + '.png')
            shutil.copy(os.path.join(curpath(), 'data', 'test.png'),
                        os.path.join(self.pngdir, fname))

        self.source = BasicSource(self.name, self.filename, self.filedate,
                                  self.temp_res, self.rootpath,
                                  self.host, self.protocol,
                                  begin_date=self.begin_date,
                                  dest_nan_value=-self.dest_nan_value,
                                  dest_regions=self.dest_regions,
                                  dest_sp_res=self.dest_sp_res,
                                  dest_temp_res=self.dest_temp_res,
                                  dest_start_date=self.dest_start_date)

    @classmethod
    def tearDownClass(self):
        if os.path.exists(self.rootpath):
            shutil.rmtree(self.rootpath)
        if os.path.exists(self.pngdir):
            shutil.rmtree(self.pngdir)

    def test_download_and_resample(self):
        self.source.download_and_resample(end=self.enddate)
        for date in self.dtindex:
            img, _, _, _ = self.source.read_img(date)
            if date == self.testdate:
                assert img[2, 4] is np.ma.masked
            else:
                assert img[2, 4] == 177.

    def test_fill_gaps(self):
        shutil.copy(os.path.join(curpath(), 'data', 'test.png'),
                    os.path.join(self.pngdir, 'test_2013_01_10.png'))
        self.source.fill_gaps(begin=self.dest_start_date, end=self.enddate)
        img, _, _, _ = self.source.read_img(self.testdate)
        assert img[1, 1] == 145.

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer thomas.mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-05-27

import calendar
from datetime import datetime, timedelta
import math

import pandas as pd
from poets.timedate.dekad import dekad_index, check_dekad


def get_dtindex(interval, begin, end=None):
    """Creates a pandas datetime index for a given interval.

    Parameters
    ----------
    interval : str or int
        Interval of the datetime index. Integer values will be treated as days.
    begin : datetime
        Datetime index start date.
    end : datetime, optional
        Datetime index end date, defaults to current date.

    Returns
    -------
    dtindex : pandas.tseries.index.DatetimeIndex
        Datetime index.
    """

    if end is None:
        end = datetime.now()

    if interval in ['dekad', 'dekadal', 'decadal', 'decade']:
        dtindex = dekad_index(begin, end)
    elif interval in ['daily', 'day', '1']:
        dtindex = pd.date_range(begin, end, freq='D')
    elif interval in ['weekly', 'week', '7']:
        begin2 = begin - timedelta(begin.weekday()) + timedelta(6)
        dtindex = pd.date_range(begin2, end, freq='7D')
    elif interval in ['monthly', 'month']:
        lday = calendar.monthrange(end.year, end.month)[1]
        end = datetime(end.year, end.month, lday)
        dtindex = pd.date_range(begin, end, freq='M')

    if type(interval) is int:
        dtindex = pd.date_range(begin, end, freq=str(str(interval) + 'D'))

    return dtindex


def check_period(interval, date):
    """
    Checks the containing interval of a date and returns the date of the
    interval.

    Parameters
    ----------
    interval : str
        Interval to check, one of (dekad, week, month).
    date : datetime
        Date to check.

    Returns
    -------
    date : datetime
        Date of the interval.
    """

    if interval in ['dekad', 'dekadal', 'decadal', 'decade']:
        date = check_dekad(date)
    elif interval in ['weekly', 'week', '7']:
        date = date - timedelta(date.weekday()) + timedelta(6)
    elif interval in ['monthly', 'month']:
        lday = calendar.monthrange(date.year, date.month)[1]
        date = datetime(date.year, date.month, lday)

    return date


def check_period_boundaries(interval, date):
    """
    Checks and returns the start and end date of the containing interval of a
    date.

    Parameters
    ----------
    interval : str
        Interval to check, one of (dekad, week, month).
    date : datetime
        Date to check.

    Returns
    -------
    begin : datetime
        Start of the interval.
    end :datetime
        End of the interval.
    """

    end = check_period(interval, date)

    if interval in ['dekad', 'dekadal', 'decadal', 'decade']:
        if end.day == calendar.monthrange(date.year, date.month)[1]:
            diff = end.day - 21
        else:
            diff = 9
        begin = end - timedelta(days=diff)
    elif interval in ['weekly', 'week', '7']:
        begin = end - timedelta(days=6)
    elif interval in ['monthly', 'month']:
        lday = calendar.monthrange(date.year, date.month)[1]
        begin = end - timedelta(days=(lday - 1))
    elif interval in ['day', 'daily']:
        begin = end

    return begin, end



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-06-03


# import pytesmo.grid.grids as grids
import pygeogrids.grids as grids
import numpy as np
import pandas as pd
import math
from shapely.geometry import Point
from poets.shape.shapes import Shape
from poets.image.imagefile import dateline_country


class RegularGrid(grids.BasicGrid):
    """Regular grid with coordinates as pixel center.

    Regular grid  that just has lat, lon coordinates and can find the nearest
    neighbour. It can also yield the gpi, lat, lon information in order.

    Parameters
    ----------
    lon : numpy.array
        longitudes of the points in the grid
    lat : numpy.array
        latitudes of the points in the grid
    gpis : numpy.array, optional
        if the gpi numbers are in a different order than the
        lon and lat arrays an array containing the gpi numbers
        can be given
        if no array is given here the lon lat arrays are given
        gpi numbers starting at 0
    subset : numpy.array, optional
        if the active part of the array is only a subset of
        all the points then the subset array which is a index
        into lon and lat can be given here
    setup_kdTree : boolean, optional
        if set (default) then the kdTree for nearest neighbour
        search will be built on initialization
    shape : tuple, optional
        if given the grid can be reshaped into the given shape
        this indicates that it is a regular grid and fills the
        attributes self.londim and self.latdim which
        define the grid only be the meridian coordinates(self.londim) and
        the coordinates of the circles of latitude(self.latdim).
        The shape has to be given as (latdim, londim)
        It it is not given the shape is set to the length of the input
        lon and lat arrays.
    sp_res : int or float
        Spatial resolution of the grid.

    Attributes
    ----------
    arrlon : numpy.array
        array of all longitudes of the grid
    arrlat : numpy.array
        array of all latitudes of the grid
    n_gpi : int
        number of gpis in the grid
    gpidirect : boolean
        if true the gpi number is equal to the index
        of arrlon and arrlat
    gpis : numpy.array
        gpi number for elements in arrlon and arrlat
        gpi[i] is located at arrlon[i],arrlat[i]
    subset : numpy.array
        if given then this contains the indices of a subset of
        the grid. This can be used if only a part of a grid is
        interesting for a application. e.g. land points, or only
        a specific country
    allpoints : boolean
        if False only a subset of the grid is active
    activearrlon : numpy.array
        array of longitudes that are active, is defined by
        arrlon[subset] if a subset is given otherwise equal to
        arrlon
    activearrlat : numpy.array
        array of latitudes that are active, is defined by
        arrlat[subset] if a subset is given otherwise equal to
        arrlat
    activegpis : numpy.array
        array of gpis that are active, is defined by
        gpis[subset] if a subset is given otherwise equal to
        gpis
    issplit : boolean
        if True then the array was split in n parts with
        the self.split function
    kdTree : object
        grid.nearest_neighbor.findGeoNN object for
        nearest neighbor search
    shape : tuple, optional
        if given during initialization then this is
        the shape the grid can be reshaped to
        this only makes sense for regular lat,lon grids
    latdim : numpy.array, optional
        if shape is given this attribute has contains
        all latitudes that make up the regular lat,lon grid
    londim : numpy.array, optional
        if shape is given this attribute has contains
        all longitudes that make up the regular lat,lon grid
    """

    def __init__(self, sp_res, **kwargs):

        lonmin = -180. + (sp_res / 2.)
        latmin = -90. + (sp_res / 2.)

        londim = np.arange(lonmin, 180, sp_res)
        latdim = np.arange(latmin, 90, sp_res)
        lon, lat = np.meshgrid(londim, latdim)

        shape = (lon.shape[1], lon.shape[0])

        super(RegularGrid, self).__init__(lon.flatten(), lat.flatten(),
                                          shape=shape, **kwargs)


class ShapeGrid(grids.BasicGrid):
    """Regular grid for a specific shape.

    Regular grid for spedific shape, that just has lat,lon coordinates and can
    find the nearest neighbour. It can also yield the gpi, lat, lon
    information in order.

    Parameters
    ----------
    lon : numpy.array
        longitudes of the points in the grid
    lat : numpy.array
        latitudes of the points in the grid
    gpis : numpy.array, optional
        if the gpi numbers are in a different order than the
        lon and lat arrays an array containing the gpi numbers
        can be given
        if no array is given here the lon lat arrays are given
        gpi numbers starting at 0
    subset : numpy.array, optional
        if the active part of the array is only a subset of
        all the points then the subset array which is a index
        into lon and lat can be given here
    setup_kdTree : boolean, optional
        if set (default) then the kdTree for nearest neighbour
        search will be built on initialization
    shape : tuple, optional
        if given the grid can be reshaped into the given shape
        this indicates that it is a regular grid and fills the
        attributes self.londim and self.latdim which
        define the grid only be the meridian coordinates(self.londim) and
        the coordinates of the circles of latitude(self.latdim).
        The shape has to be given as (latdim, londim)
        It it is not given the shape is set to the length of the input
        lon and lat arrays.
    region : str, optional
        Identifier of the region in the shapefile. If the default shapefile is
        used, this would be the FIPS country code.
    sp_res : float
        Spatial resolution of the grid.
    shapefile : str, optional
        Path to shape file, uses "world country admin boundary shapefile" by
        default.

    Attributes
    ----------
    arrlon : numpy.array
        array of all longitudes of the grid
    arrlat : numpy.array
        array of all latitudes of the grid
    n_gpi : int
        number of gpis in the grid
    gpidirect : boolean
        if true the gpi number is equal to the index
        of arrlon and arrlat
    gpis : numpy.array
        gpi number for elements in arrlon and arrlat
        gpi[i] is located at arrlon[i],arrlat[i]
    subset : numpy.array
        if given then this contains the indices of a subset of
        the grid. This can be used if only a part of a grid is
        interesting for a application. e.g. land points, or only
        a specific country
    allpoints : boolean
        if False only a subset of the grid is active
    activearrlon : numpy.array
        array of longitudes that are active, is defined by
        arrlon[subset] if a subset is given otherwise equal to
        arrlon
    activearrlat : numpy.array
        array of latitudes that are active, is defined by
        arrlat[subset] if a subset is given otherwise equal to
        arrlat
    activegpis : numpy.array
        array of gpis that are active, is defined by
        gpis[subset] if a subset is given otherwise equal to
        gpis
    issplit : boolean
        if True then the array was split in n parts with
        the self.split function
    kdTree : object
        grid.nearest_neighbor.findGeoNN object for
        nearest neighbor search
    shape : tuple, optional
        if given during initialization then this is
        the shape the grid can be reshaped to
        this only makes sense for regular lat,lon grids
    latdim : numpy.array, optional
        if shape is given this attribute has contains
        all latitudes that make up the regular lat,lon grid
    londim : numpy.array, optional
        if shape is given this attribute has contains
        all longitudes that make up the regular lat,lon grid
    region : str
        Identifier of the region in the shapefile. If the default shapefile is
        used, this would be the FIPS country code.
    sp_res : float
        spatial resolution of the grid
    shp : poets.shape.shapes.Country
        Information about the country/region shape
    """

    def __init__(self, region, sp_res, shapefile=None):

        self.country = region

        self.shp = Shape(region, shapefile=shapefile)

        # countries that cross the international dateline (maybe more!)
        if region in ['NZ', 'RS', 'US']:
            lonmin, lonmax = dateline_country(region)
        else:
            lonmin, lonmax = minmaxcoord(self.shp.bbox[0], self.shp.bbox[2],
                                         sp_res)

        latmin, latmax = minmaxcoord(self.shp.bbox[1], self.shp.bbox[3],
                                     sp_res)

        if region in ['NZ', 'US', 'RS']:
            lons1 = np.arange(lonmin, 180, sp_res)
            lons2 = (np.arange(-180, lonmax + sp_res, sp_res))
            lons = np.empty(len(lons1) + len(lons2))
            lons[0:len(lons1)] = lons1[:]
            lons[len(lons1):] = lons2[:]
        else:
            lons = np.arange(lonmin, lonmax + sp_res, sp_res)

        lats = np.arange(latmin, latmax + sp_res, sp_res)

        lon_new, lat_new = _remove_blank_frame(region, lons, lats, shapefile)

        lon, lat = np.meshgrid(lon_new, lat_new)

        shape = (lon.shape[1], lon.shape[0])

        super(ShapeGrid, self).__init__(lon.flatten(), lat.flatten(),
                                        shape=shape)

    def get_gridpoints(self):
        """Gets all points within a country shape.

        Removes all gridpoints that are within country bounding box, but
        outside of the country border.

        Returns
        -------
        points : pd.DataFrame
            Dataframe containing all points with GPI as index, lat and lon as
            columns.
        """

        box = self.get_bbox_grid_points(self.shp.bbox[1], self.shp.bbox[3],
                                        self.shp.bbox[0], self.shp.bbox[2])
        poly = self.shp.polygon  # MultiPolygon

        lons = []
        lats = []
        pts = []

        for gpi in box:
            lon, lat = self.gpi2lonlat(gpi)
            if poly.contains(Point(lon, lat)):
                    pts.append(gpi)
                    lons.append(lon)
                    lats.append(lat)

        points = pd.DataFrame({'lon': lons, 'lat': lats}, pts)

        return points


def _remove_blank_frame(region, lons, lats, shapefile=None):
    """Removes longitudes and latitudes without points in region shape.

    Parameters
    ----------
    region : str
        FIPS country code (https://en.wikipedia.org/wiki/FIPS_country_code)
    lons : numpy.ndarray
        Array with longitudes.
    lats : numpy.ndarray
        Array with latitudes.

    Returns
    --------
    lon_new : list of floats
        Updated list of longitudes.
    lat_new : list of floats
        Updated list of latitudes.
    """

    shp = Shape(region, shapefile)

    poly = shp.polygon

    del_lons = []
    del_lats = []

    # left boundary check
    for i, x in enumerate(lons[:(lons.size / 2)]):
        checksum = 0
        for y in lats:
            p = Point(x, y)
            if not p.within(poly):
                checksum += 1
        if checksum == lats.size:
            del_lons.append(i)
        else:
            break

    # right boundary check
    for i, x in enumerate(lons[::-1][:(lons.size / 2)]):
        checksum = 0
        for y in lats:
            p = Point(x, y)
            if not p.within(poly):
                checksum += 1
        if checksum == lats.size:
            del_lons.append(lons.size - 1 - i)
        else:
            break

    # bottom boundary check
    for i, y in enumerate(lats[:(lats.size / 2)]):
        checksum = 0
        for x in lons:
            p = Point(x, y)
            if not p.within(poly):
                checksum += 1
        if checksum == lons.size:
            del_lats.append(i)
        else:
            break

    # top boundary check
    for i, y in enumerate(lats[::-1][:(lats.size / 2)]):
        checksum = 0
        for x in lons:
            p = Point(x, y)
            if not p.within(poly):
                checksum += 1
        if checksum == lons.size:
            del_lats.append(lats.size - 1 - i)
        else:
            break

    lon_new = lons.tolist()
    lat_new = lats.tolist()

    for i in del_lons:
        if lons[i] in lon_new:
            lon_new.remove(lons[i])

    for i in del_lats:
        if lats[i] in lat_new:
            lat_new.remove(lats[i])

    return lon_new, lat_new


def minmaxcoord(min_threshold, max_threshold, sp_res):
    """Gets min and max coordinates of a specific grid.

    Based on the frame of a global grid.

    Parameters
    ----------
    min_threshold : float
        Minimum value of coordinate
    max_threshold : float
        Maximum value of coordinate
    sp_res : float or int
        Spatial resolution or the grid

    Returns
    -------
    minval : float
        Updated minimum coordinate
    maxval : float
        Updated maximum coordinate
    """

    res = float(sp_res)

    minval = int(math.ceil(min_threshold / res)) * res
    maxval = int(math.floor(max_threshold / res)) * res

    if minval != maxval:
        if minval - (res / 2) < min_threshold:
            minval += res / 2
        else:
            minval -= res / 2

        if maxval + (res / 2) > max_threshold:
            maxval -= res / 2
        else:
            maxval += res / 2

    return minval, maxval



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer
# Creation date: 2014-07-29

"""
This module includes the poets base class `Poet`.
"""

from datetime import datetime
from netCDF4 import Dataset
from poets.grid.grids import ShapeGrid
from poets.io.source_base import BasicSource
import numpy as np
import os
import pandas as pd
import poets.web.app as app

valid_temp_res = ['dekad', 'dekade', 'month', 'monthly', 'week', 'weekly',
                  'day', 'daily']


class Poet(object):
    """POETS base class.

    Provides methods to download and resample data using parameters as defined
    in this class. Resampled outputfiles will be saved as NetCDF4 files.

    Parameters
    ----------
    rootpath : str
        path to the directory where data should be stored
    regions : list of str, str, optional
        Identifier of the region in the shapefile. If the default shapefile is
        used, this would be the FIPS country code. Defaults to global.
    region_names : list of str, optional
        Full name of the regions. If set, must have same size as regions
        parameter. Defaults to regions parameter.
    spatial_resolution : float, optional
        spatial resolution in degree, defaults to 0.25
    temporal_resolution : str, optional
        temporal resolution of the data, possible values: day, week, month,
        dekad; defaults to dekad.
    start_date : datetime.datetime, optional
        first date of the dataset, defaults to 2000-01-01
    nan_value : int
        NaN value to use, defaults to -99
    shapefile : str, optional
        Path to shape file, uses "world country admin boundary shapefile" by
        default. Custom shapefile must use WGS84 as reference system.
    delete_rawdata : bool, optional
        Original files will be deleted from rawdata_path if set True. Defaults
        to False.

    Attributes
    ----------
    rootpath : str
        path to the directory where data should be stored
    regions : list of str
        Identifier of the region in the shapefile.
    region_names : list of str
        Full name of the regions.
    spatial_resolution : float
        Spatial resolution in degree.
    temporal_resolution : str
        Temporal resolution of the data.
    data_path : str
        Path where resampled NetCDF file is stored.
    rawdata_path : str
        Path where original files are stored and downloaded.
    tmp_path : str
        Path where temporary files are stored.
    nan_value : int
        NaN value to use, defaults to -99.
    start_date : datetime.datetime
        First date of the dataset.
    shapefile : str
        Path to shape file.
    sources : dict of poets.io.BasicSource objects
        Sources used by poets given as BasicSource class.
    delete_rawdata : bool
        Original files will be deleted from rawdata_path if True.
    """

    def __init__(self, rootpath, regions=['global'],
                 spatial_resolution=0.25, temporal_resolution='dekad',
                 start_date=datetime(2000, 1, 1), nan_value=-99, url=None,
                 shapefile=None, delete_rawdata=False, region_names=None):

        self.rootpath = rootpath
        if isinstance(regions, str):
            self.regions = [regions]
        else:
            self.regions = regions
        if region_names is not None:
            if isinstance(region_names, str):
                self.region_names = [region_names]
            else:
                self.region_names = region_names
        else:
            self.region_names = self.regions
        self.spatial_resolution = spatial_resolution

        if temporal_resolution not in valid_temp_res:
            raise ValueError("Temporal resulution must be one of " +
                             str(valid_temp_res))

        self.temporal_resolution = temporal_resolution
        self.rawdata_path = os.path.join(rootpath, 'RAWDATA')
        self.data_path = os.path.join(rootpath, 'DATA')
        self.tmp_path = os.path.join(rootpath, 'TMP')
        self.nan_value = nan_value
        self.start_date = start_date
        self.shapefile = shapefile
        self.delete_rawdata = delete_rawdata
        self.sources = {}
        self.url = url

        if not os.path.exists(self.rawdata_path):
            os.mkdir(self.rawdata_path)
        if not os.path.exists(self.tmp_path):
            os.mkdir(self.tmp_path)
        if not os.path.exists(self.data_path):
            os.mkdir(self.data_path)

    def add_source(self, name, filename, filedate, temp_res, host, protocol,
                   username=None, password=None, port=22, directory=None,
                   dirstruct=None, begin_date=None, regions=None,
                   variables=None, nan_value=None,
                   valid_range=None, unit=None, ffilter=None, data_range=None,
                   colorbar=None, src_file=None, labels=None, xticks=None):
        """Creates BasicSource class and adds it to `Poet.sources`.

        Parameters
        ----------
        name : str
            Name of the data source.
        filename : str
            Structure/convention of the file name.
        filedate : dict
            Position of date fields in filename, given as tuple.
        temp_res : str
            Temporal resolution of the source.
        host : str
            Link to data host.
        protocol : str
            Protocol for data transfer.
        username : str, optional
            Username for data access.
        password : str, optional
            Password for data access.
        port : int, optional
            Port to data host, defaults to 22.
        directory : str, optional
            Path to data on host.
        dirstruct : list of strings
            Structure of source directory, each list item represents a
            subdirectory.
        regions : list of str, optional
            List of regions where data from source is available. Uses all
            regions as given in Poet.regions attribute if None.
        begin_date : datetime.date, optional
            Date from which on data is available.
        variables : string or list of strings, optional
            Variables used from data source.
        nan_value : int, float, optional
            Nan value of the original data as given by the data provider.
        valid_range : tuple of int of float, optional
            Valid range of data, given as (minimum, maximum).
        data_range : tuple of int of float, optional
            Range of the values as data given in rawdata (minimum, maximum).
            Will be scaled to valid_range.
        ffilter : str, optional
            Pattern that apperas in filename. Can be used to select out not
            needed files if multiple files per date are provided.
        colorbar : str, optional
            Colorbar to use, use one from
            http://matplotlib.org/examples/color/colormaps_reference.html;
            defaults to jet.
        labels : list, optional
            Custom tick-labels for the legend in the web-app; must have same
            dimension as xticks and only works if xticks is set; Defaults
            to None.
        xticks : list of int or float, optional
            Custom tick locations for the legend in the web-app; must have same
            dimension as labels and only works if labels is set; Defaults
            to None.
        unit : str, optional
            Unit of dataset for displaying in legend. Does not have to be set
            if unit is specified in input file metadata. Defaults to None.
        src_file : dict of str, optional
            Path to file that contains source. Uses default NetCDF file if
            None.
        """

        source = BasicSource(name, filename, filedate, temp_res, self.rootpath,
                             host, protocol, username=username,
                             password=password, port=port, ffilter=ffilter,
                             directory=directory, dirstruct=dirstruct,
                             begin_date=begin_date, variables=variables,
                             nan_value=nan_value, valid_range=valid_range,
                             unit=unit, regions=regions,
                             data_range=data_range, colorbar=colorbar,
                             dest_nan_value=self.nan_value,
                             dest_regions=self.regions,
                             dest_sp_res=self.spatial_resolution,
                             dest_temp_res=self.temporal_resolution,
                             dest_start_date=self.start_date,
                             src_file=src_file, labels=labels, xticks=xticks)

        self.sources[name] = source

    def fetch_data(self, begin=None, end=None, delete_rawdata=None,
                   source=None):
        """Starts download and resampling of input data for sources as added
        to `Poets.sources`.

        Parameters
        ----------
        begin : datetime, optional
            Start date of data to download, defaults to start date as defined
            in poets class.
        end : datetime, optional
            End date of data to download, defaults to current datetime.
        delete_rawdata : bool, optional
            Original files will be deleted from rawdata_path if set True.
            Defaults to value of delete_rawdata attribute as set in Poet class.
        source = str, optional
            Name of the source to fetch_data. If None, all sources will be
            fetched.
        """

        if delete_rawdata is None:
            delete_rawdata = self.delete_rawdata

        for src in self.sources.keys():
            if source is not None:
                if source != src:
                    continue
            s = self.sources[src]
            print '[INFO] Download data for source ' + src
            s.download_and_resample(begin=begin, end=end,
                                    shapefile=self.shapefile,
                                    delete_rawdata=delete_rawdata)

        print '[SUCCESS] Download and resampling complete!'

    def download(self, begin=None, end=None, source=None):
        """Starts download of input data for sources as added
        to `Poets.sources`.

        Parameters
        ----------
        begin : datetime, optional
            Start date of data to download, defaults to start date as defined
            in poets class.
        end : datetime, optional
            End date of data to download, defaults to current datetime.
        source = str, optional
            Name of the source download. If None, all sources will be
            downloaded.
        """

        for src in self.sources.keys():
            if source is not None:
                if source != src:
                    continue
            s = self.sources[src]
            print '[INFO] Download data for source ' + src
            s.download(begin=begin, end=end)

        print '[SUCCESS] Download complete!'

    def resample(self, begin=None, end=None, delete_rawdata=None, source=None):
        """Starts download of input data for sources as added
        to `Poets.sources`.

        Parameters
        ----------
        begin : datetime, optional
            Start date of data to download, defaults to start date as defined
            in poets class.
        end : datetime, optional
            End date of data to download, defaults to current datetime.
        source = str, optional
            Name of the source to resample. If None, all sources will be
            resampled.
        """

        for src in self.sources.keys():
            if source is not None:
                if source is not src:
                    continue
            s = self.sources[src]
            print '[INFO] Resampling data for source ' + src
            s.resample(begin=begin, end=end,
                       shapefile=self.shapefile,
                       delete_rawdata=delete_rawdata)

        print '[SUCCESS] Resampling complete!'

    def fill_gaps(self, source=None):
        """
        Detects gaps in data and tries to fill them by downloading and
        resampling the data within these periods.
        source = str, optional
            Name of the source to fill gaps. If None, all sources will be
            fixed.
        """

        for src in self.sources.keys():
            if source is not None:
                if source is not src:
                    continue
            s = self.sources[src]
            print '[INFO] Scanning ' + src + ' for gaps'
            s.fill_gaps()

    def get_gridpoints(self):
        """Returns gridpoints from NetCDF file.

        Parameters
        ----------
        region : str
            Identifier of the region in the NetCDF file.

        Returns
        -------
        gridpoints : dict of pandas.DataFrame
            Dict containing Dataframes with gridpoint index as index,
            longitutes and latitudes as columns for each region.
        """

        gridpoints = {}

        if self.regions == ['global']:
            filename = (self.regions[0] + '_' + str(self.spatial_resolution)
                        + '_' + str(self.temporal_resolution) + '.nc')
            ncfile = os.path.join(self.data_path, filename)

            with Dataset(ncfile, 'r+', format='NETCDF4') as nc:
                gpis = nc.variables['gpi'][:]
                lons = nc.variables['lon'][:]
                lats = nc.variables['lat'][:]
                gpis = gpis.flatten()
                lons, lats = np.meshgrid(lons, lats)
                lons = lons.flatten()
                lats = lats.flatten()

            points = pd.DataFrame(index=gpis)
            points['lon'] = lons
            points['lat'] = lats
            gridpoints['global'] = points
        else:
            for region in self.regions:
                grid = ShapeGrid(region, self.spatial_resolution,
                                 self.shapefile)
                points = grid.get_gridpoints()
                gridpoints[region] = points

        return gridpoints

    def read_image(self, source, date, region=None, variable=None):
        """Gets images from netCDF file for certain date

        Parameters
        ----------
        date : datetime
            Date of the image.
        source : str
            Data source from which image should be read.
        region : str, optional
            Region of interest, set to first defined region if None.
        variable : str, optional
            Variable to display, set to first variable of source if None.

        Returns
        -------
        img : numpy.ndarray
            Image of selected date.
        lon : numpy.array
            Array with longitudes.
        lat : numpy.array
            Array with latitudes.
        metadata : dict
            Dictionary containing metadata of the variable.
        """

        img, lon, lat, metadata = self.sources[source].read_img(date, region,
                                                                variable)

        return img, lon, lat, metadata

    def read_timeseries(self, source, location, region=None, variable=None):
        """
        Gets timeseries from netCDF file for a gridpoint.

        Parameters
        ----------
        source : str
            Data source from which time series should be read.
        location : int or tuple of floats
            Either Grid point index as integer value or Longitude/Latitude
            given as tuple.
        region : str, optional
            Region of interest, set to first defined region if None.
        variable : str, optional
            Variable to display, set to first variable of source if None.

        Returns
        -------
        ts : pd.DataFrame
            Timeseries for the selected data.
        """

        ts = self.sources[source].read_ts(location, region, variable,
                                          shapefile=self.shapefile)

        return ts

    def get_variables(self, region=None):
        """
        Returns all variables available.

        Parameters
        ----------
        region : str, optional
            Region to check for variables.

        Returns
        -------
        variables : list of str
            Sorted list of all variables.
        """

        variables = []

        for src in self.sources.keys():
            if region is None:
                for var in self.sources[src].get_variables():
                    variables.append(var)
            else:
                if self.sources[src].regions is not None:
                    if region in self.sources[src].regions:
                        for var in self.sources[src].get_variables():
                            variables.append(var)
                else:
                    for var in self.sources[src].get_variables():
                        variables.append(var)

        variables.sort()

        return variables

    def start_app(self, host='127.0.0.1', port=None, r_host=None, r_port=None,
                  url=None, debug=False):
        """Starts web interface.

        Parameters
        ----------
        host : str, optional
            Host that is used by the app, defaults to 127.0.0.1.
        port : int, optional
            Port where app runs on, defaults to None.
        r_host : str, optional
            IP of router that is between host and internet.
        r_port : int, optional
            Port of router that is between host and internet.
        url : str, optional
            URL where poets web interface is hosted. If not set, poets will run
            on the IP of the hosting machine.
        debug : bool, optional
            Starts app in debug mode if set True, defaults to False.
        """

        app.start(self, host, port, r_host, r_port, url, debug)




# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-10-02

import os

supported_formats = ['.nc', '.nc4',
                     '.h5',
                     '.tiff', '.tif',
                     '.png', '.jpg', '.jpeg', '.gif']


def check_supported(filename):
    """Checks if file is in supported format.

    Parameters
    ----------
    filename : str
        Filename or filepath.

    Returns
    -------
    bool
        True if supported, False if not.
    """

    return os.path.splitext(filename)[1].lower() in supported_formats


def select_file(filelist):
    """Selects a file out of a list of files, based on their extension.

    Parameters
    ----------
    filelist : list of str
        List containing filepaths.

    Returns
    -------
    filename : str
        Filepath of selected file.

    Raises
    ------
    IOError :
        If filelist contains no supported file format.
    """

    extlist = []

    for f in filelist:
        fext = os.path.splitext(f)[1].lower()
        if fext in supported_formats:
            extlist.append(supported_formats.index(fext))
        else:
            extlist.append('X')

    val, idx = min((val, idx) for (idx, val) in enumerate(extlist))

    if val == 'X':
        raise IOError("No valid file format found.")
    else:
        return filelist[idx]




# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer thomas.mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-08-13

import unittest
from datetime import datetime
from poets.timedate.dateindex import get_dtindex, check_period


class Test(unittest.TestCase):

    def setUp(self):
        self.begin = datetime(2004, 2, 1)
        self.end = datetime(2004, 3, 31)

    def tearDown(self):
        pass

    def test_get_dtindex(self):

        interval1 = 'dekad'
        interval2 = 'day'
        interval3 = 'week'
        interval4 = 'month'
        interval5 = 8

        dtindex1 = get_dtindex(interval1, self.begin, self.end)
        dtindex2 = get_dtindex(interval2, self.begin, self.end)
        dtindex3 = get_dtindex(interval3, self.begin, self.end)
        dtindex4 = get_dtindex(interval4, self.begin, self.end)
        dtindex5 = get_dtindex(interval5, self.begin, self.end)

        assert dtindex1.size == 6
        assert dtindex1[0] == datetime(2004, 2, 10)
        assert dtindex1[-1] == datetime(2004, 3, 31)

        assert dtindex2.size == 60
        assert dtindex2[0] == datetime(2004, 2, 1)
        assert dtindex2[-1] == datetime(2004, 3, 31)

        assert dtindex3.size == 9
        assert dtindex3[0] == datetime(2004, 2, 1)
        assert dtindex3[-1] == datetime(2004, 3, 28)

        assert dtindex4.size == 2
        assert dtindex4[0] == datetime(2004, 2, 29)
        assert dtindex4[-1] == datetime(2004, 3, 31)

        assert dtindex5.size == 8
        assert dtindex5[0] == datetime(2004, 2, 1)
        assert dtindex5[-1] == datetime(2004, 3, 28)

    def test_check_period(self):
        dat = datetime(2014, 11, 20)
        assert check_period('dekad', dat) == datetime(2014, 11, 20, 0, 0)
        assert check_period('week', dat) == datetime(2014, 11, 23, 0, 0)
        assert check_period('monthly', dat) == datetime(2014, 11, 30, 0, 0)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-10-01

'''
Module for unpacking compressed archives. Based on pyunpack and patool.
'''

import os
import shutil
from pyunpack import Archive


def check_compressed(filepath):
    """
    Checks if a file is compressed using the file extension.

    Parameters
    ----------
    filepath : string
        Path to input file.

    Returns
    -------
    boolean
        True if compressed, False if not.
    """

    comp_ext = ['.zip', '.bz2', '.tar', '.7z', '.ace', '.rar', '.adf', '.alz',
                '.cab', '.Z', '.cpio', '.deb', '.dms', '.gz', '.iso', '.lrz',
                '.lha', '.lzh', '.lz', '.lzma', '.lzo', '.rpm', '.rz', '.shn',
                '.xz', '.zoo']

    if os.path.splitext(filepath)[1].lower() in comp_ext:
        return True
    else:
        return False


def flatten(outpath):
    '''
    Flattens directory structure.

    Parameters
    ----------
    outpath : str
        Directory to flatten.

    Raises
    ------
    OSError :
        If file cannot be moved.
    '''

    # move files
    for dirpath, dirnames, filenames in os.walk(outpath):
        for filename in filenames:
            try:
                os.rename(os.path.join(dirpath, filename),
                          os.path.join(outpath, filename))
            except OSError:
                print ("Could not move %s " % os.path.join(dirpath,
                                                           filename))
    # delete empty folders
    for dirpath, dirnames, filenames in os.walk(outpath):
        for directory in dirnames:
            shutil.rmtree(os.path.join(dirpath, directory))


def unpack(filepath, outpath=None):
    """
    Unpacks compressed archives and files recursively and flattens the output.

    Parameters
    ----------
    filepath : str
        Path to zipped archive.
    outpath : str
        Path where decompressed files will be stored.
    flatten : bool, optional
        If True, output dir will be flattened.

    Raises
    ------
    IOError :
        If input file does not exist.
    """

    if os.path.isfile(filepath):

        if outpath is None:
            outpath = os.path.splitext(filepath)[0]
            outpath = outpath.split(".")[0]

        if not os.path.exists(outpath):
            os.makedirs(outpath)

        Archive(filepath).extractall(outpath)

        flatten(outpath)

        for f in os.listdir(outpath):
            if check_compressed(f):
                unpack(os.path.join(outpath, f))
                os.unlink(os.path.join(outpath, f))

        flatten(outpath)

    else:
        raise IOError('%s file does not exist' % filepath)




# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Isabella Pfeil, isy.pfeil@gmx.at
# Creation date: 2014-11-24

import unittest
import os
import poets.io.unpack as up
import zipfile
import shutil


def curpath():
    pth, _ = os.path.split(os.path.abspath(__file__))

    return pth


class Test(unittest.TestCase):

    def setUp(self):
        self.fname = os.path.join(curpath(), 'data', 'test.zip')
        self.outpath = os.path.join(curpath(), 'data', 'test')
        self.dirpath = os.path.join(curpath(), 'data', 'testdir')

    def tearDown(self):

        if os.path.exists(self.dirpath):
            shutil.rmtree(self.dirpath)
        if os.path.exists(os.path.join(curpath(), 'data', 'test2.txt')):
            os.remove(os.path.join(curpath(), 'data', 'test2.txt'))

    def test_check_compressed(self):
        test = up.check_compressed(self.fname)
        self.failUnless(test is True)

    def test_unpack(self):
        # create txt-file
        txt1 = open(os.path.join(curpath(), 'data', 'test1.txt'), 'w')
        txt1.close()

        # create first zip archive
        ziparc = zipfile.ZipFile(os.path.join(curpath(), 'data',
                                              'ziparc.zip'), 'w')

        # write txt-file and folder to zip archive
        ziparc.write(os.path.join(curpath(), 'data', 'test1.txt'))
        ziparc.close()

        # 'create' second txt-file
        os.rename(os.path.join(curpath(), 'data', 'test1.txt'),
                  os.path.join(curpath(), 'data', 'test2.txt'))

        # create second zip archive
        ziparc2 = zipfile.ZipFile(os.path.join(curpath(), 'data',
                                               'ziparc2.zip'), 'w')

        # write second txt-file and first zip archive to second zip archive
        ziparc2.write(os.path.join(curpath(), 'data', 'test2.txt'))
        ziparc2.write(os.path.join(curpath(), 'data', 'ziparc.zip'))
        ziparc2.close()

        dirlen = 2
        up.unpack(ziparc2.filename, self.outpath)

        unpack_len = len(os.listdir(self.outpath))

        assert unpack_len == dirlen

        shutil.rmtree(self.outpath)
        os.remove(os.path.join(curpath(), 'data', ziparc.filename))
        os.remove(os.path.join(curpath(), 'data', ziparc2.filename))

    def test_flatten(self):
        if not os.path.exists(self.dirpath):
            os.makedirs(self.dirpath)
            os.makedirs(os.path.join(self.dirpath, 'dir1'))
            os.makedirs(os.path.join(self.dirpath, 'dir2'))
            self.fn1 = open(os.path.join(self.dirpath, 'dir1', 'fn1.txt'), 'a')
            self.fn2 = open(os.path.join(self.dirpath, 'dir1', 'fn2.txt'), 'a')
            self.fn3 = open(os.path.join(self.dirpath, 'dir2', 'fn3.txt'), 'a')
            self.fn1.close()
            self.fn2.close()
            self.fn3.close()

        dirlen = 3
        up.flatten(self.dirpath)

        flatten_len = len(os.listdir(self.dirpath))

        assert flatten_len == dirlen

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Isabella Pfeil, isy.pfeil@gmx.at
# Creation date: 2014-07-18

"""
This module provides functions for converting lonlat-information
to pixels in an image file (eg. PNG, TIFF) and for calculating a
country bounding box.
"""

import numpy as np
import math
from osgeo import gdal
from PIL import Image
from poets.shape.shapes import Shape
from poets.image.geotiff import lonlat2px_gt, px2lonlat_gt


def get_layer_extent(filepath):
    """
    Returns extent of imagefile as minimum and maximum longitudes and 
    latitudes.

    Parameters:
    -----------
    filepath : str
        Path to image file.

    Returns:
    --------
    lon_min : float
        Minimum longitude.
    lat_min : float
        Minimum latitude.
    lon_max : float
        Maximum longitude.
    lat_max : float
        Maximum latitude.
    """
    ds = gdal.Open(filepath)
    gt = ds.GetGeoTransform()
    width = ds.RasterXSize
    height = ds.RasterYSize
    lon_min = gt[0]
    lat_min = gt[3] + width * gt[4] + height * gt[5]
    lon_max = gt[0] + width * gt[1] + height * gt[2]
    lat_max = gt[3]

    return lon_min, lat_min, lon_max, lat_max


def lonlat2px(img, lon, lat):
    """
    Converts a pair of lon and lat to its corresponding pixel value in an
    image file.

    Parameters
    ----------
    img : PIL image
        Input image file, e.g. PNG, TIFF.
    lon : float
        Longitude.
    lat : float
        Latitude.

    Returns
    -------
    Row : float
        Corresponding pixel value.
    Col : float
        Corresponding pixel value.
    """

    w, h = img.size

    mw = w / 360.0
    mh = h / 180.0

    row = h / 2 - lat * mh
    col = w / 2 + lon * mw

    return row, col


def lonlat2px_rearr(img, lon, lat):
    """
    Converts a pair of lon and lat to its corresponding pixel
    value in a rearranged image file (see rearrange_img).

    Parameters
    ----------
    img : PIL image
        Input image file, e.g. PNG, TIFF.
    lon : float
        Longitude
    lat : float
        Latitude

    Returns
    -------
    Row : float
        corresponding pixel value
    Col : float
        corresponding pixel value
    """

    w, h = img.size
    mw = w / 360.0
    mh = h / 180.0

    row = h / 2 - lat * mh

    if lon >= 0:
        col = 0 + lon * mw
    elif lon < 0:
        col = w + lon * mw

    return row, col


def px2lonlat(img, lon_px, lat_px):
    """
    Converts two arrays of row and column pixels into their corresponding lon 
    and lat arrays.

    Parameters
    ----------
    img : PIL image
        Input image file, e.g. PNG, TIFF.
    lon_px : np.array
        Array of column pixels.
    lat_px : np.array
        Array of row pixels.

    Returns
    -------
    lon_new : np.array
        List of corresponding longitude values.
    lat_new: np.array
        List of corresponding latitude values.
    """

    w, h = img.size

    mw = w / 360.0
    mh = h / 180.0

    lon_new = np.zeros(len(lon_px))
    lat_new = np.zeros(len(lat_px))

    for i in range(0, len(lon_px)):
        lon_new[i] = (lon_px[i] - w / 2) / mw

    for i in range(0, len(lat_px)):
        lat_new[i] = -(lat_px[i] - h / 2) / mh

    return lon_new, lat_new


def px2lonlat_rearr(img, lon_px, lat_px):
    """
    Converts two arrays of row and column pixels into their
    corresponding lon and lat arrays

    Parameters
    ----------
    img : PIL image
        Image which the pixel values refer to (rearranged image)
    lon_px : np.array
        Array of column pixels.
    lat_px : np.array
        Array of row pixels.

    Returns
    -------
    lon_new : np.array
        List of corresponding longitude values.
    lat_new: np.array
        List of corresponding latitude values.
    """

    w, h = img.size
    mw = w / 360.0
    mh = h / 180.0

    lon_new = np.zeros(len(lon_px))
    lat_new = np.zeros(len(lat_px))

    for i in range(0, len(lon_px)):  # lon [-180, 179.999]
        if lon_px[i] >= w / 2:  # west
            lon_new[i] = -(w - lon_px[i]) / mw
        elif lon_px[i] < w / 2:  # east
            lon_new[i] = lon_px[i] / mw

    for i in range(0, len(lat_px)):
        lat_new[i] = -(lat_px[i] - h / 2) / mh

    return lon_new, lat_new


def rearrange_img(img):
    """
    Rearranges image so that 0 degree Meridian is on the very left.
    Used when area around the +- 180 degree Meridian is of interest
    (eastern Russia, Alaska, New Zealand...).

    Parameters
    ----------
    img : PIL image
        Image to be rearranged.

    Returns
    -------
    img2 : Image file
        Rearranged image.
    """

    w, h = img.size
    blocklen = w / 2
    xblock = w / blocklen
    yblock = 1
    blockmap = [(xb * blocklen, yb * blocklen, (xb + 1) * blocklen,
                 (yb + 1) * blocklen) for xb in xrange(xblock)
                for yb in xrange(yblock)]
    rearr = blockmap[::-1]
    img2 = Image.new(img.mode, (w, h))
    for box, sbox in zip(blockmap, rearr):
        c = img.crop(sbox)
        img2.paste(c, box)

    return img2


def dateline_country(country):
    """
    Min and max longitude for countries that spread across the international 
    dateline.

    Returns
    -------
    lon_min : float
        Minimum longitude.
    lon_max : float
        Maximum longitude.
    """

    if country == 'NZ':
        lon_min = 165.0 + 52.0 / 60.0 + 12.0 / 3600.0
        lon_max = -(175.0 + 50.0 / 60.0)

    elif country == 'US':
        lon_min = 173.0 + 11.0 / 60.0
        lon_max = -(66.0 + 59.0 / 60.0 + 0.71006 / 3600.0)

    elif country == 'RS':
        lon_min = 19.0 + 38.0 / 60.0
        lon_max = -(169.0 + 3.0 / 60.0 + 54.0 / 3600.0)

    return lon_min, lon_max


def bbox_img(source_file, region, fileExtension, shapefile=None):
    """
    Clips bounding box out of image file and returns data as numpy.ndarray

    Parameters
    ----------
    source_file : str
        Path to source file.
    region : str
        Identifier of the region in the shapefile. If the default shapefile is
        used, this would be the FIPS country code.
    fileExtension : str
        Filetype (e.g. png, tif).
    shapefile : str, optional
        Path to shape file, uses "world country admin boundary shapefile" by
        default.

    Returns
    -------
    data : dict of numpy.arrays
        Clipped image (grey values).
    lon_new : numpy.array
        Longitudes of the clipped image.
    lat_new : numpy.array
        Latitudes of the clipped image.
    timestamp : datetime.date
        Timestamp of the image.
    metadata : dict of strings
        Metadata from source netCDF file.
    """

    orig_img = Image.open(source_file)

    lon_min_src, lat_min_src, lon_max_src, lat_max_src = \
        get_layer_extent(source_file)

    if region == 'global':
        lon_min = -180
        lon_max = 180
        lat_min = -90
        lat_max = 90
    else:
        shp = Shape(region, shapefile)
        lon_min = shp.bbox[0]
        lon_max = shp.bbox[2]
        lat_min = shp.bbox[1]
        lat_max = shp.bbox[3]

    d = lon_max - lon_min

    # countries that cross the international dateline (maybe more!)
    if region in ['NZ', 'RS', 'US']:
        lon_min, lon_max = dateline_country(region)

    # get 2 pairs of points (upper left, lower right of bbox)
    if d > 350 and region not in ['AY', 'global']:
        if fileExtension in ['.tif', '.tiff', '.TIF', '.TIFF']:
            if (round(lon_max_src - lon_min_src) == 360.0 and
                round(lat_max_src - lat_min_src) == 180.0):
                orig_img = rearrange_img(orig_img)
                row_min, col_min = lonlat2px_rearr(orig_img, lon_min, lat_max)
                row_max, col_max = lonlat2px_rearr(orig_img, lon_max, lat_min)

                img = orig_img.crop((int(math.floor(col_min)),
                                     int(math.floor(row_min)),
                                     int(math.ceil(col_max)),
                                     int(math.ceil(row_max))))
            else:
                print 'Rearranging is only possible for global imagefiles.'
                return
        else:
            orig_img = rearrange_img(orig_img)
            row_min, col_min = lonlat2px_rearr(orig_img, lon_min, lat_max)
            row_max, col_max = lonlat2px_rearr(orig_img, lon_max, lat_min)

            img = orig_img.crop((int(math.floor(col_min)),
                                 int(math.floor(row_min)),
                                 int(math.ceil(col_max)),
                                 int(math.ceil(row_max))))

    elif fileExtension in ['.tif', '.tiff', '.TIF', '.TIFF']:
        row_min, col_min = lonlat2px_gt(orig_img, lon_min, lat_max,
                                        lon_min_src, lat_min_src, lon_max_src,
                                        lat_max_src)
        row_max, col_max = lonlat2px_gt(orig_img, lon_max, lat_min,
                                        lon_min_src, lat_min_src, lon_max_src,
                                        lat_max_src)

        # crop image
        img = orig_img.crop((int(math.floor(col_min)),
                             int(math.floor(row_min)),
                             int(math.ceil(col_max)),
                             int(math.ceil(row_max))))
    else:
        row_min, col_min = lonlat2px(orig_img, lon_min, lat_max)
        row_max, col_max = lonlat2px(orig_img, lon_max, lat_min)

        # crop image
        img = orig_img.crop((int(math.floor(col_min)),
                             int(math.floor(row_min)),
                             int(math.ceil(col_max)),
                             int(math.ceil(row_max))))

    # get data values from image
    data = {'dataset': np.array(img)}

    # lon_new, lat_new
    lon_px = np.arange(int(math.floor(col_min)), int(math.ceil(col_max)))
    lat_px = np.arange(int(math.floor(row_min)), int(math.ceil(row_max)))

    if region in ['NZ', 'RS', 'US']:
        lon_new, lat_new = px2lonlat_rearr(orig_img, lon_px, lat_px)
    elif fileExtension in ['.tif', '.tiff', '.TIF', '.TIFF']:
        lon_new, lat_new = px2lonlat_gt(orig_img, lon_px, lat_px, lon_min_src,
                                        lat_min_src, lon_max_src, lat_max_src)
    else:
        lon_new, lat_new = px2lonlat(orig_img, lon_px, lat_px)

    # timestamp
    timestamp = None

    # metadata
    metadata = None

    # move coordinates to pixel center
    if lon_new.size > 2:
        lon_new += (lon_new[1] - lon_new[0]) / 2
    if lat_new.size > 2:
        lat_new += (lat_new[1] - lat_new[0]) / 2

    return data, lon_new, lat_new, timestamp, metadata



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer Thomas.Mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-07-07

import unittest
from poets.grid.grids import ShapeGrid, RegularGrid


class Test(unittest.TestCase):

    def setUp(self):
        self.region = 'AU'
        self.sp_res = 0.25
        self.sp_res1 = 0.5

    def tearDown(self):
        pass

    def test_ShapeGrid(self):
        # general case
        cgrid = ShapeGrid(self.region, sp_res=self.sp_res)
        bbox = (cgrid.arrlat.min(), cgrid.arrlat.max(),
                cgrid.arrlon.min(), cgrid.arrlon.max())

        assert cgrid.get_grid_points()[0].size == 290
        assert bbox == (46.625, 48.875, 9.875, 16.875)
        assert cgrid.get_gridpoints().shape == (158, 2)

        # test special case NZ
        cgrid = ShapeGrid('NZ', sp_res=self.sp_res1)
        bbox = (cgrid.arrlat[0], cgrid.arrlat[-1],
                cgrid.arrlon[0], cgrid.arrlon[-1])

        assert cgrid.shape == (34, 24)
        assert cgrid.get_gridpoints().shape == (116, 2)
        assert bbox == (-46.75, -35.25, 167.37, -176.5)

    def test_RegularGrid(self):
        points_number = 259200
        bbox = (-89.75, 89.75, -179.75, 179.75)
        grid_shape = (720, 360)

        grid = RegularGrid(sp_res=self.sp_res1)

        grid_bbox = (grid.arrlat.min(), grid.arrlat.max(),
                     grid.arrlon.min(), grid.arrlon.max())

        assert grid.get_grid_points()[0].size == points_number
        assert grid_bbox == bbox
        assert grid.shape == grid_shape

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()
